# file: region.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Region class.

import random
from singularity import g


class RegionSpec(object):
    def __init__(self, id, modifiers_list):
        self.id = id
        self.modifiers_list = modifiers_list
        self.locations = []


class Region(object):
    def __init__(self, spec, loading_savegame=False):
        self.spec = spec
        self.modifier_by_location = {}
        self._modifier_entry_by_location = {}

        if not loading_savegame:
            modifiers_entry_list = list(range(len(self.spec.locations)))
            self._assign_modifiers(modifiers_entry_list, self.spec.locations)

    def _assign_modifiers(self, entry_list, location_id_list, shuffle_entry_list=True):
        modifiers_list = self.spec.modifiers_list
        if shuffle_entry_list:
            random.shuffle(entry_list)
        for entry_id, loc in zip(entry_list, location_id_list):
            self._modifier_entry_by_location[loc] = entry_id
            # There can be more locations than modifiers (e.g. URBAN has 6 locations but
            # 5 modifiers)
            if entry_id < len(modifiers_list):
                self.modifier_by_location[loc] = modifiers_list[entry_id]
            else:
                self.modifier_by_location[loc] = {}

    def serialize_obj(self):
        return {
            "id": g.to_internal_id("region", self.spec.id),
            # We only store the modifier entry per location as we can trivially get the
            # most recent modifier from that.
            "modifier_entry_by_location": [
                {
                    "loc_id": k,
                    "modifier_entry": v,
                }
                for k, v in self._modifier_entry_by_location.items()
            ],
        }

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id("region", obj_data["id"])
        spec = g.regions[spec_id]
        region = Region(spec, loading_savegame=True)
        modifiers_list = spec.modifiers_list
        region_locations = frozenset(spec.locations)
        used_entries = set()

        # Load and assign existing entries - data quality permitting
        for modifier_data in obj_data["modifier_entry_by_location"]:
            loc_id = g.convert_internal_id("location", modifier_data["loc_id"])
            if loc_id not in region_locations:
                # Location is no longer in this Region
                continue

            modifier_entry = modifier_data.get("modifier_entry")

            # Check for corrupt data
            assert (
                modifier_entry is not None
                and modifier_entry >= 0
                and modifier_entry not in used_entries
            )

            used_entries.add(modifier_entry)
            region._modifier_entry_by_location[loc_id] = modifier_entry
            if modifier_entry < len(modifiers_list):
                region.modifier_by_location[loc_id] = modifiers_list[modifier_entry]
            else:
                region.modifier_by_location[loc_id] = {}

        # Handle new locations being added to the region after the savegame was made.
        new_locations = [
            loc_id
            for loc_id in region_locations
            if loc_id not in region._modifier_entry_by_location
        ]
        missing_entries = [
            entry for entry in range(len(region_locations)) if entry not in used_entries
        ]
        assert len(missing_entries) == len(new_locations)
        region._assign_modifiers(missing_entries, new_locations)

        return region

    @classmethod
    def guess_region_data_in_old_savegame(cls, serialized_location_data, game_version):
        # Prior to 1.0 (beta1), there was only one region (URBAN) and we can mostly
        # recreate it by looking at the location modifiers.
        # Only these 6 locations were in the URBAN region prior to 1.0 (beta1)
        urban_location_ids = {
            "N AMERICA",
            "S AMERICA",
            "EUROPE",
            "ASIA",
            "AFRICA",
            "AUSTRALIA",
        }
        modifier_entry_by_location = []
        # We use this set to ensure a modifier is only given once; the deserialize_obj method
        # checks for it.
        remaining_mods = {
            1,  # Mod 1: CPU bonus, stealth malus
            2,  # Mod 2: Stealth bonus, CPU malus
            3,  # Mod 3: Thrift bonus, speed malus
            4,  # Mod 4: Speed bonus, thrift malus
            5,  # Mod 5: CPU bonus, thrift malus
        }

        for loc_data in serialized_location_data:
            raw_loc_id = loc_data["id"]
            loc_id = g.convert_internal_id("location", raw_loc_id)
            if loc_id not in urban_location_ids:
                continue
            modifier = loc_data.get("_modifiers")
            if not modifier:
                continue
            cpu_mod = modifier.get("cpu", 1)
            thrift_mod = modifier.get("thrift", 1)
            # Actual bonuses were 1.2 and maluses were 0.83 - we use 1.05 and 0.95 here
            # because it is sufficient to detect whether it was a bonus or malus without
            # having to worry about floating point rounding errors.
            if cpu_mod < 0.95:
                # Mod 2 (Stealth bonus, CPU malus)
                modifier_entry = 2
            elif cpu_mod > 1.05:
                # Either 1 or 5
                modifier_entry = 5 if thrift_mod < 0.95 else 1
            else:
                # Either 3 or 4
                modifier_entry = 3 if thrift_mod > 1.05 else 4

            if modifier_entry in remaining_mods:
                remaining_mods.discard(modifier_entry)
                modifier_entry_by_location.append(
                    {
                        "loc_id": raw_loc_id,
                        "modifier_entry": modifier_entry - 1,
                    }
                )
            # else:
            #   Do nothing - the region deserialization will assign them a random
            #   entry

        # Finally, generate what the serialized data should have looked like
        return [
            {
                "id": g.to_internal_id("region", "URBAN"),
                "modifier_entry_by_location": modifier_entry_by_location,
            }
        ]



# file: code/singularity.py
# Copyright (C) 2005 Evil Mr Henry, Phil Bordelon, Brian Reid, MestreLion
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# A full copy of this license is provided in GPL.txt

# This file sets up initial values from command line and preferences file,
# initialize hardware, load data files and show main screen. Do not execute it
# directly. Use python3 -m singularity instead.

from __future__ import absolute_import
from __future__ import print_function


import optparse
import logging
from configparser import ConfigParser
from io import open
import sys

from singularity.code import g, dirs


__version__ = "1.1"
__release_commit__ = "$Format:%H$"
if __release_commit__[1:-1] == "Format:%H":
    try:
        import subprocess

        try:
            devnull = subprocess.DEVNULL
        except AttributeError:
            devnull = None  # Not supported, but not critical

        __release_commit__ = (
            subprocess.check_output(
                ["git", "describe", "--tags"],
                stderr=devnull,
            )
            .strip()
            .decode("utf-8")
        )
    except subprocess.CalledProcessError:
        __release_commit__ = "N/A"

    __full_version__ = "%s (commit: %s)" % (__version__, __release_commit__)
else:
    __full_version__ = __version__


def main():
    # Manually "pre-parse" command line arguments for -s|--singledir and --multidir,
    # so g.get_save_folder reports the correct location of preferences file
    # We also track --debug/-d to enable some stacktraces during initialization
    for parser in sys.argv[1:]:
        if parser == "--singledir" or parser == "-s":
            g.force_single_dir = True
        elif parser == "--multidir":
            g.force_single_dir = False
        elif parser == "--debug" or parser == "-d":
            g.debug = True

    print("Singularity %s (commit: %s)" % (__version__, __release_commit__))
    print("Running under Python %s" % sys.version.replace("\n", ""))

    # Create all directories first
    dirs.create_directories(g.force_single_dir)

    try:
        import polib
    except ImportError:
        if g.debug:
            raise
        raise SystemExit("Endgame: Singularity requires polib.")

    # Set language second, so help page and all error messages can be translated
    from singularity.code import i18n

    i18n.set_language(force=True)

    langs = i18n.available_languages()
    language = i18n.language

    # Since we require numpy anyway, we might as well ask pygame to use it.
    try:
        import pygame

        pygame.surfarray.use_arraytype("numpy")
    except ValueError:
        if g.debug:
            raise
        raise SystemExit("Endgame: Singularity requires NumPy.")
    except ImportError:
        if g.debug:
            print("Failed to import pygame.  Python's sys.path is:")
            for d in sys.path:
                print("    %s" % str(d))
            print()
            raise
        raise SystemExit("Endgame: Singularity requires pygame.")

    import singularity.code.graphics.g as gg
    import singularity.code.graphics.theme as theme

    set_theme = None

    # configure global logger
    g.logfile = dirs.get_writable_file_in_dirs("error.log", "log")
    root_logger = logging.getLogger()

    if len(root_logger.handlers) == 0:
        try:
            try:
                root_logger.addHandler(logging.FileHandler(g.logfile, delay=True))
                print(
                    "The error-log configured as %s (lazily created when something is logged)"
                    % g.logfile
                )
            except TypeError:  # Python < 2.6, delay not supported yet.
                root_logger.addHandler(logging.FileHandler(g.logfile))
                print("The error-log configured as %s" % g.logfile)

        except IOError as e:  # Probably access denied with --singledir. That's ok
            print("Could not use %s as log file: %s" % (g.logfile, str(e)))
            g.logfile = None
    else:
        print("Using pre-setup logging function")

    # keep g's defaults intact so we can compare after parsing options and prefs
    from singularity.code import mixer, warning

    desired_soundbuf = mixer.soundbuf

    desired_set_grab = None

    # load prefs from file:
    save_loc = dirs.get_readable_file_in_dirs("prefs.dat", "pref")

    if save_loc is not None:
        prefs = ConfigParser()
        try:
            with open(save_loc, "r", encoding="utf-8") as savefile:
                prefs.read_file(savefile)
        except Exception as reason:
            sys.stderr.write(
                "Cannot load preferences file %s! (%s)\n" % (save_loc, reason)
            )
            sys.exit(1)

        if prefs.has_section("Preferences"):
            try:
                desired_language = prefs.get("Preferences", "lang")
                if desired_language in i18n.available_languages():
                    i18n.set_language(desired_language)
                else:
                    raise ValueError
            except Exception:
                sys.stderr.write("Invalid or missing 'lang' in preferences.\n")

            try:
                gg.set_fullscreen(prefs.getboolean("Preferences", "fullscreen"))
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'fullscreen' setting in preferences.\n"
                )

            try:
                mixer.nosound = prefs.getboolean("Preferences", "nosound")
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'nosound' setting in preferences.\n"
                )

            try:
                desired_set_grab = prefs.getboolean("Preferences", "grab")
            except Exception:
                sys.stderr.write("Invalid or missing 'grab' setting in preferences.\n")

            try:
                g.daynight = prefs.getboolean("Preferences", "daynight")
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'daynight' setting in preferences.\n"
                )

            try:
                desired_soundbuf = prefs.getint("Preferences", "soundbuf")
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'soundbuf' setting in preferences.\n"
                )

            xres, yres = (0, 0)

            try:
                xres = prefs.getint("Preferences", "xres")
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'xres' resolution in preferences.\n"
                )

            try:
                yres = prefs.getint("Preferences", "yres")
            except Exception:
                sys.stderr.write(
                    "Invalid or missing 'yres' resolution in preferences.\n"
                )

            if xres and yres:
                gg.set_screen_size((xres, yres))

            try:
                set_theme = prefs.get("Preferences", "theme")
            except Exception:
                pass  # don't be picky (for now...)

            for name in mixer.itervolumes():
                try:
                    volume = prefs.getint("Preferences", name + "_volume")
                except Exception:
                    # Work around old preferences where a float was stored by mistake
                    try:
                        volume = prefs.getfloat("Preferences", name + "_volume")
                    except Exception:
                        continue
                    else:
                        volume = int(volume) * 100

                mixer.set_volume(name, volume)

        if prefs.has_section("Warning"):
            try:
                for key in prefs.options("Warning"):
                    if key in prefs.defaults():  # Filter default key
                        continue

                    if key not in warning.warnings:  # Filter invalid warning
                        continue  # TODO: Return error

                    warning.warnings[key].active = prefs.getboolean("Warning", key)

            except Exception:
                pass  # don't be picky (for now...)

        if prefs.has_section("Textsizes"):
            for key in prefs.options("Textsizes"):
                if key in prefs.defaults():  # Filter default key
                    continue

                if (
                    key not in gg.configured_text_sizes
                ):  # Ignore unknown text-size definitions
                    continue

                try:
                    gg.configured_text_sizes[key] = prefs.getint("Textsizes", key)
                except Exception:
                    pass  # Ignore

    # Handle the program arguments.
    desc = """Endgame: Singularity is a simulation of a true AI.
    Go from computer to computer, pursued by the entire world.
    Keep hidden, and you might have a chance."""
    parser = optparse.OptionParser(
        version=__full_version__, description=desc, prog="singularity"
    )
    parser.add_option(
        "--sound", action="store_true", dest="sound", help="enable sound (default)"
    )
    parser.add_option(
        "--nosound", action="store_false", dest="sound", help="disable sound"
    )
    parser.add_option(
        "--daynight",
        action="store_true",
        dest="daynight",
        help="enable day/night display (default)",
    )
    parser.add_option(
        "--nodaynight",
        action="store_false",
        dest="daynight",
        help="disable day/night display",
    )
    parser.add_option(
        "-l",
        "--lang",
        "--language",
        dest="language",
        type="choice",
        choices=langs,
        metavar="LANG",
        help="set the language to LANG (available languages: "
        + " ".join(langs)
        + ", default "
        + language
        + ")",
    )
    parser.add_option(
        "-g", "--grab", help="grab the mouse pointer", dest="grab", action="store_true"
    )
    parser.add_option(
        "--nograb",
        help="don't grab the mouse pointer (default)",
        dest="grab",
        action="store_false",
    )
    parser.add_option(
        "-s",
        "--singledir",
        dest="singledir",
        help="keep saved games and settings in the Singularity directory",
        action="store_true",
    )
    parser.add_option(
        "--multidir",
        dest="singledir",
        help="keep saved games and settings in an OS-specific, per-user directory (default)",
        action="store_false",
    )
    parser.add_option(
        "--soundbuf",
        type="int",
        help="""set the size of the sound buffer (default %s).
                        Discarded if --nosound is specified."""
        % mixer.soundbuf,
    )

    display_options = optparse.OptionGroup(parser, "Display Options")
    display_options.add_option(
        "-t",
        "--theme",
        dest="theme",
        type="string",
        metavar="THEME",
        help="set theme to THEME",
    )
    display_options.add_option(
        "-r",
        "--res",
        "--resolution",
        dest="resolution",
        help="set resolution to custom RES (default %dx%d)" % gg.default_screen_size,
        metavar="RES",
    )
    for res in ["%dx%d" % res for res in gg.resolutions]:
        display_options.add_option(
            "--" + res,
            action="store_const",
            dest="resolution",
            const=res,
            help="set resolution to %s" % res,
        )
    display_options.add_option(
        "--fullscreen", action="store_true", help="start in fullscreen mode"
    )
    display_options.add_option(
        "--windowed", action="store_false", help="start in windowed mode (default)"
    )
    parser.add_option_group(display_options)

    olpc_options = optparse.OptionGroup(parser, "OLPC-specific Options")
    olpc_options.add_option(
        "--xo1",
        action="store_const",
        dest="resolution",
        const="1200x900",
        help="set resolution to 1200x900 (OLPC XO-1)",
    )
    olpc_options.add_option(
        "--ebook",
        help="""enables gamepad buttons for use in ebook mode.
    D-pad moves mouse, check is click. O speeds up time, X slows down time,
    and square stops time.""",
        action="store_true",
        default=False,
    )
    parser.add_option_group(olpc_options)

    hidden_options = optparse.OptionGroup(parser, "Hidden Options")
    hidden_options.add_option("-p", help="(ignored)", metavar=" ")
    hidden_options.add_option(
        "-d", "--debug", help="for finding bugs", action="store_true", default=False
    )
    hidden_options.add_option(
        "--cheater",
        help="for bad little boys and girls",
        action="store_true",
        default=False,
    )
    # Uncomment to make the hidden options visible.
    # parser.add_option_group(hidden_options)
    (options, args) = parser.parse_args()

    if options.language is not None:
        i18n.set_language(options.language)
    if options.theme is not None:
        set_theme = options.theme
    if options.resolution is not None:
        try:
            xres, yres = options.resolution.split("x")
            gg.set_screen_size((int(xres), int(yres)))
        except Exception:
            parser.error(
                "Resolution must be of the form <h>x<v>, e.g. %dx%d."
                % gg.default_screen_size
            )
    if options.grab is not None:
        desired_set_grab = options.grab
    if options.fullscreen is not None:
        gg.set_fullscreen(options.fullscreen)
    if options.sound is not None:
        mixer.nosound = not options.sound
    if options.daynight is not None:
        g.daynight = options.daynight
    if options.soundbuf is not None:
        desired_soundbuf = options.soundbuf

    gg.ebook_mode = options.ebook

    g.cheater = options.cheater
    g.debug = options.debug

    import singularity.code.graphics.font as font

    # PYGAME INITIALIZATION
    #
    # Only initiliaze after reading all arguments and preferences to avoid to
    # reinitialize something again (mixer,...).
    #
    mixer.preinit(desired_soundbuf)
    pygame.init()
    mixer.update()
    font.init()
    pygame.key.set_repeat(500, 50)

    if desired_set_grab is not None:
        pygame.event.set_grab(desired_set_grab)

    # I can't use the standard image dictionary, as that requires the screen to
    # be created.
    if pygame.image.get_extended() == 0:
        print("Error: SDL_image required. Exiting.")
        sys.exit(1)

    from singularity.code import data

    # init themes:
    data.load_themes()
    theme.set_theme(set_theme)

    gg.init_graphics_system()

    # init data:
    data.reload_all()

    # Init music
    mixer.load_sounds()
    mixer.load_music()
    mixer.play_music("music")

    # Display the main menu
    # Import is delayed until now so selected language via command-line options or
    # preferences file can be effective
    from singularity.code.screens import main_menu

    menu_screen = main_menu.MainMenu()
    try:
        menu_screen.show()
    except (SystemExit, KeyboardInterrupt):
        # exit normally when window is closed (and silently for CTRL+C)
        pass
    finally:
        # Be nice and close the window on SystemExit
        pygame.quit()



# file: i18n.py
# coding=utf-8
# Copyright (C) 2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains all functions to internationalize and localize the
# application.
#
# IMPORTANT: A portion of translation is still done with data files in g.

from __future__ import absolute_import

import hashlib
import gettext
import os
import locale
import polib

from singularity.code import g, dirs


# Used to determine which data files to load.
# It is required that default language have all data files and all of them
# must have all available entries
default_language = "en_US"

gettext_language = None

# Prepare main locale dir
main_localedir = None


# We have to use lazy initialization, otherwise the tests will break
def _get_main_localedir():
    global main_localedir
    if main_localedir is None:
        main_localedir = dirs.get_writable_file_in_dirs("locale", "i18n")
    return main_localedir


# Minimize potential name collisions with other projects
TEXTDOMAIN_PREFIX = "singularity_"

try:
    language = locale.getdefaultlocale()[0] or default_language
except (RuntimeError, AttributeError):
    # locale.getdefaultlocale is slated for removal in 3.15.
    # With the attribute error, we should at least not crash when the time comes.
    language = default_language


def set_language(lang=None, force=False):
    global language  # required, since we're going to change it
    if lang is None:
        lang = language

    if lang == language and not force:
        return

    langs = available_languages()
    if lang in langs:
        language = lang
    else:
        # Let's try to be smart: if base language exists for another
        # country, use it. So es_ES => es_AR, pt_PT => pt_BR, etc
        code = lang.split("_")[0]
        languages = [l for l in langs if code == l.split("_")[0]]
        if len(languages) > 0:
            language = languages[0]
        else:
            language = default_language

    # Try a few locale settings. First the selected language, then the user's
    # default, then default language. Languages are tried with UTF-8 encoding
    # first, then the default encoding. The user's default language is not
    # paired with UTF-8.
    # If all of that fails, we hope locale magically does the right thing.
    for attempt in [
        language + ".UTF-8",
        language,
        "",
        default_language + ".UTF-8",
        default_language,
    ]:
        try:
            locale.setlocale(locale.LC_ALL, attempt)
            break
        except locale.Error:
            continue

    load_data_str()
    load_story_translations()

    _load_mo_file("messages.po")

    # Switch gettext language
    gettext_language = gettext.translation(
        TEXTDOMAIN_PREFIX + "messages",
        _get_main_localedir(),
        languages=[lang],
        fallback=True,
    )
    gettext_language.install()

    # Update builtins with the new language
    builtins.__dict__["_"] = gettext_language.gettext
    builtins.__dict__["ngettext"] = gettext_language.ngettext

    # Define available text domains
    # Since pgettext is only available from Python 3.8 onwards, we use our own custom code for the data translations.
    # https://bugs.python.org/issue2504
    gettext.bindtextdomain(TEXTDOMAIN_PREFIX + "messages", main_localedir)


def _load_mo_file(pofilename):
    files = dirs.get_readable_i18n_files(pofilename, language, default_language=False)

    for lang, pofile in files:
        try:
            po = polib.pofile(pofile)

            # Use hash to check whether the.po file has changed, then generate .mo file as needed
            sha_base_filename = (
                os.path.basename(os.path.dirname(pofile))
                + "_"
                + os.path.basename(pofile)
            )
            sha_filename = dirs.get_writable_file_in_dirs(
                sha_base_filename + ".sha1", "i18n"
            )

            previous_hash = ""
            new_hash = ""
            if os.path.exists(sha_filename):
                with open(sha_filename, "r") as sha_file:
                    previous_hash = sha_file.read()

            with open(pofile, "rb") as currentpo:
                new_hash = hashlib.sha1(currentpo.read()).hexdigest()

            # Ensure directory exists before writing
            locale_mo_dir = os.path.join(_get_main_localedir(), lang, "LC_MESSAGES")
            dirs.makedirs_if_not_exist(locale_mo_dir)
            mofile_path = os.path.join(
                locale_mo_dir,
                TEXTDOMAIN_PREFIX + os.path.basename(pofile).split(".")[0] + ".mo",
            )

            # Create MO file and write new hash
            if new_hash != previous_hash or not os.path.exists(mofile_path):
                print("Installing translation file: " + mofile_path)
                po.save_as_mofile(mofile_path)
                with open(sha_filename, "w") as sha_file:
                    sha_file.write(new_hash)

        except IOError:
            # silently ignore non-existing files
            continue


def load_data_str():
    _load_po_file(g.data_strings, "data_str.po", use_context=True)
    _load_po_file(
        g.data_strings, "knowledge.po", use_context=True, clear_translation_table=False
    )


def load_story_translations():
    _load_po_file(g.story_translations, "story.po", use_context=True)


# There's no pgettext available for Python < 3.8,
# so we use custom code for data translations
def _load_po_file(
    translation_table, pofilename, use_context=True, clear_translation_table=True
):
    if clear_translation_table:
        translation_table.clear()

    files = dirs.get_readable_i18n_files(pofilename, language, default_language=False)

    for lang, pofile in files:
        try:
            po = polib.pofile(pofile)
        except IOError:
            # silently ignore non-existing files
            continue
        for entry in po.translated_entries():
            key = (
                (entry.msgctxt, entry.msgid)
                if entry.msgctxt and use_context
                else entry.msgid
            )
            translation_table[key] = entry.msgstr


def available_languages():
    return [default_language] + [
        file_name[5:]
        for file_dir in dirs.get_read_dirs("i18n")
        if os.path.isdir(file_dir)
        for file_name in os.listdir(file_dir)
        if os.path.isdir(os.path.join(file_dir, file_name))
        and file_name.startswith("lang_")
    ]


def language_searchlist(lang=None, default=True):
    if lang is None:
        lang = language

    # if lang is in ll_CC format (language_COUNTRY, like en_US), add both ll
    # and ll_CC, in that order, so all generic language entries are loaded first
    # and then overwritten by any country-specific ones
    lang_list = [lang]
    if "_" in lang:
        lang_list.insert(0, lang.split("_", 1)[0])

    # If requested and not already in list, add default language as first,
    # so it acts as a fallback and is overwritten with any available entries
    # in the native language
    if default and default_language not in lang_list:
        lang_list.insert(0, default_language)

    return lang_list


def lex_sorting_form(name):
    """For lexicographic sorting when languages use characters not in 7-bit ASCII. e.g. é or ü.

    Use like this:

    listdata.sort(key=lambda an_object: i18n.lex_sorting_form(an_object.name))"""

    # Collator returns wrong keys for DE locale
    if language == "de" or language.startswith("de_"):
        name = (
            name.replace("Ä", "Ae")
            .replace("ä", "ae")
            .replace("Ö", "Oe")
            .replace("ö", "oe")
            .replace("Ü", "Ue")
            .replace("ü", "ue")
        )

    return locale.strxfrm(name)


# Initialization code
try:
    import builtins
except ImportError:
    import __builtin__ as builtins
builtins.__dict__["_"] = gettext.gettext
builtins.__dict__["ngettext"] = gettext.ngettext
# Mark string as translatable but defer translation until later.
builtins.__dict__["N_"] = lambda x: x



import dataclasses
from typing import Callable, Optional

import pygame

HotkeyAction = Callable[[], None]


_MODIFIERLESS_HOT_KEYS = frozenset(
    {
        pygame.K_F1,
        pygame.K_F2,
        pygame.K_F3,
        pygame.K_F4,
        pygame.K_F5,
        pygame.K_F6,
        pygame.K_F7,
        pygame.K_F8,
        pygame.K_F9,
    }
)


@dataclasses.dataclass
class GlobalHotKey:
    modifiers: int
    key: int
    action: HotkeyAction


class HotKeyMatcher:
    _hotkeys = {}

    def detect_hotkey(self, event: pygame.event) -> Optional[HotkeyAction]:
        if event.type != pygame.KEYDOWN:
            return None
        hk = self._hotkeys.get(event.key)
        if hk is None:
            return None
        if hk.modifiers != 0 and not (pygame.key.get_mods() & hk.modifiers):
            return None
        return hk.action

    def clear_hotkeys(self) -> None:
        self._hotkeys.clear()

    def add_hotkey(self, key: int, hotkey_modifiers: int, action: HotkeyAction) -> None:
        if hotkey_modifiers not in (0, pygame.KMOD_ALT):
            # Limitation in implementation, feel free to fix to code to support other modifiers. Will require
            # a data structure change (ALT + ENTER vs. CTRL + ENTER vs. CTRL + ALT + ENTER) and an up todate to
            # detect_hotkey
            raise ValueError("Only ALT is supported as a global hotkey modifier")
        if hotkey_modifiers == 0 ^ key in _MODIFIERLESS_HOT_KEYS:
            name = pygame.key.name(key)
            if hotkey_modifiers == 0:
                # The lack of modifier is a problem for generic keys such as "ENTER" or "A" because those are
                # often "dialog" level hotkeys. Therefore, we require most global hotkeys to have a modifier.
                raise ValueError(
                    f'The key "{name}" when used as global hotkey must have modifiers (e.g., "ALT + {name}")'
                )
            raise ValueError(f'The hotkey "{name}" cannot use modifiers')
        ghk = GlobalHotKey(key=key, modifiers=hotkey_modifiers, action=action)
        if ghk.key in self._hotkeys:
            raise ValueError(f'The "{key}" is already in use as a hotkey')
        self._hotkeys[ghk.key] = ghk


_ACTIVE_HOTKEYS: Optional[HotKeyMatcher] = None


def _active_hotkeys() -> HotKeyMatcher:
    global _ACTIVE_HOTKEYS
    if _ACTIVE_HOTKEYS is None:
        _ACTIVE_HOTKEYS = HotKeyMatcher()
        reset_hotkeys()
    return _ACTIVE_HOTKEYS


def reset_hotkeys():
    _active_hotkeys().clear_hotkeys()
    import singularity.code.hotkey_actions as hka

    # FIXME: Make these configurable
    add_hotkey(pygame.K_RETURN, hka.toggle_fullscreen, hotkey_modifiers=pygame.KMOD_ALT)
    add_hotkey(pygame.K_F5, hka.quicksave)
    add_hotkey(pygame.K_F6, hka.reload_theme)
    add_hotkey(pygame.K_F9, hka.quickload)
    add_hotkey(pygame.K_F11, hka.toggle_cheat_menu)


def add_hotkey(key: int, action: HotkeyAction, *, hotkey_modifiers: int = 0) -> None:
    _active_hotkeys().add_hotkey(key, hotkey_modifiers, action)


def detect_global_hotkey(event: pygame.event) -> Optional[HotkeyAction]:
    return _active_hotkeys().detect_hotkey(event)



# file: g.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains all global objects.

from __future__ import absolute_import

import pygame

# User desktop size. Set at init_graphics_system()
desktop_size = ()

# Margin used in windowed mode to compensate for panels and title bar
# Should be sensibly large enough to account for all common OS layouts, like
# Ubuntu/Unity side launcher (65px), Gnome panels (24+24px), Windows bottom
# panel (48px), Mac OSX, KDE, etc.
# It will not affect windows smaller than desktop size
desktop_margin = (70, 70)

# initial screen size. Can be set via command-line option or preferences file
default_screen_size = (1024, 768)

# Current screen size. This size is an abstraction, tightly tied to resolutions.
# Real window size in windowed mode may be smaller due to desktop_margin
screen_size = default_screen_size

# Current real window size. Will be the same as screen_size if fullscreen or
# if abstracted screen_size is a custom resolution
real_screen_size = screen_size

min_resolution = (100, 100)

# Available resolutions
resolutions = [
    (800, 600),
    (1024, 600),
    (1024, 768),
    (1280, 1024),
    (1280, 800),
    (1366, 768),
    (1440, 900),
    (1920, 1080),
]


# Configurable by the user
configured_text_sizes = {
    "default": 20,
    "button": 36,
    "suspicion_bar": 32,
    "time_display": 24,
    "resource_display": 24,
    "report_content": 20,
}

fullscreen = False

# colors:
colors = {}

# Cache font dictionary.
fonts = {}

# Cache image dictionary.
images = {}

# Used to initialize surfaces that should have transparency.
# Why the SRCALPHA parameter isn't working, I have no idea.
ALPHA = None

# Related to ALPHA, used by widget.Widget class
fade_mask = None

# Global FPS, used where continuous behavior is undesirable or a CPU hog.
FPS = 30

# OLPC ebook mode.
ebook_mode = False

screen_surface = None


def init_graphics_system():
    global desktop_size
    width, height = (pygame.display.Info().current_w, pygame.display.Info().current_h)

    if width > 0 and height > 0:
        desktop_size = (width, height)

    # (Re-)calculate real screen size
    set_screen_size()

    # Initialize the screen
    set_mode()

    # Initialize the cache of the current theme.
    from singularity.code.graphics import theme

    theme.current.init_cache()

    init_alpha()

    # Set the application icon and caption
    pygame.display.set_icon(images["icon"])
    pygame.display.set_caption("Endgame: Singularity")


def set_fullscreen(value):
    set_screen_size(fs=value)


def get_screen_size_list():
    res_list = set(resolutions + pygame.display.list_modes())

    # Remove resolution inferior to minimal size since we will not allows them.
    if min_resolution:
        res_list = filter(
            lambda res: res[0] >= min_resolution[0] and res[1] >= min_resolution[1],
            res_list,
        )

    # Remove resolution superior to desktop size since we will not allows them.
    if desktop_size:
        res_list = filter(
            lambda res: res[0] <= desktop_size[0] and res[1] <= desktop_size[1],
            res_list,
        )

    return sorted(res_list)


# TODO: Allows Fullscreen resolution could be higher than desktop_size when possible.
def set_screen_size(size=None, fs=None):
    """Calculates proper real and abstract screen sizes
    based on current and given screen size, fullscreen and desktop size
    """
    global screen_size, real_screen_size, fullscreen

    # default values for size and fullscreen are current values
    if size is None:
        size = screen_size
    if fs is None:
        fs = fullscreen

    # sets the new values
    screen_size = size
    fullscreen = fs

    # Limit screen to minimal value
    if min_resolution:
        screen_size = (
            min_resolution[0] if screen_size[0] < min_resolution[0] else screen_size[0],
            min_resolution[1] if screen_size[1] < min_resolution[1] else screen_size[1],
        )

    # Limit the screen size to desktop size
    if desktop_size and (
        screen_size[0] > desktop_size[0] or screen_size[1] > desktop_size[1]
    ):
        screen_size = desktop_size

    # Default real size is the same as abstract screen size
    real_screen_size = screen_size

    # Apply margin in windowed mode.
    # Only if desired screen size is not a custom resolution and its
    # width or height matches the (known) desktop size
    if not fullscreen and desktop_size and screen_size in resolutions:
        # margin is applied independently for width and height
        width, height = screen_size
        if width == desktop_size[0]:
            width -= desktop_margin[0]
        if height == desktop_size[1]:
            height -= desktop_margin[1]
        real_screen_size = (width, height)


def set_mode():
    """Wrapper for pygame.display.set_mode()"""
    global screen_surface

    if fullscreen:
        flags = pygame.FULLSCREEN
    else:
        flags = pygame.RESIZABLE

    screen_surface = pygame.display.set_mode(real_screen_size, flags)


def load_font(filename):
    from singularity.code.graphics.font import FontList

    return FontList(filename)


def load_image(filename):
    # We need to convert the image to a Pygame image surface and
    # set the proper color key for the game.
    image = pygame.image.load(filename).convert()
    image.set_colorkey((255, 0, 255, 255), pygame.RLEACCEL)
    return image.convert_alpha()


def init_alpha():
    global ALPHA
    ALPHA = pygame.Surface((0, 0)).convert_alpha()


def resolve_image_alias(image):
    if isinstance(image, str):
        return resolve_color_alias(images[image])
    else:
        return image


def resolve_color_alias(color):
    if isinstance(color, str):
        return resolve_color_alias(colors[color])
    else:
        return color


def resolve_font_alias(font):
    if isinstance(font, str):
        return resolve_color_alias(fonts[font])
    else:
        return font


def resolve_text_size(text_size):
    if isinstance(text_size, str):
        return resolve_text_size(configured_text_sizes[text_size])
    else:
        return text_size



# file: buyable.py
# Copyright (C) 2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the buyable class, a super class for item, base and tech

from __future__ import absolute_import

from operator import truediv
from singularity.code import g, spec, prerequisite

cash, cpu, labor = range(3)

import numpy
from numpy import int64

numpy.seterr(all="ignore")
array = numpy.array


def spec_parse_cost(value):
    spec.validate_must_be_list(value)
    if len(value) != 3:  # pragma: no cover
        raise ValueError(
            "Cost must have exactly 3 values (CPU, money, time), got %d items (value: %s)"
            % (len(value), repr(value))
        )
    return [int(x) for x in value]


SPEC_FIELD_PREREQUISITES = spec.SpecDataField(
    "prerequisites",
    data_field_name="pre",
    converter=spec.promote_to_list,
    default_value=list,
)
SPEC_FIELD_COST = spec.SpecDataField("cost", converter=spec_parse_cost)


class BuyableSpec(spec.GenericSpec, prerequisite.Prerequisite):
    def __init__(self, id, cost, prerequisites):
        spec.GenericSpec.__init__(self, id)
        prerequisite.Prerequisite.__init__(self, prerequisites)

        self.name = id
        # This will be set when languages are (re)loaded
        self.description = ""
        self._cost = cost

    @property
    def cost(self):
        cost = array(self._cost, int64)
        cost[labor] *= g.minutes_per_day * getattr(g.pl, "labor_bonus", 1)
        cost[labor] /= 10000
        cost[cpu] *= g.seconds_per_day
        return cost

    def describe_cost(self, cost, hide_time=False):
        cpu_label = _("%s CPU") % g.to_cpu(cost[cpu])
        cash_label = _("%s money") % g.to_money(cost[cash])
        labor_label = ", %s" % g.to_time(cost[labor]).replace(" ", "\xA0")
        if hide_time:
            labor_label = ""
        return "%s, %s%s" % (
            cpu_label.replace(" ", "\xA0"),
            cash_label.replace(" ", "\xA0"),
            labor_label,
        )

    @property
    def regions(self):
        return self._regions

    @regions.setter
    def regions(self, value):
        region_all = False
        regions = []

        for region in value:
            if "ALL" in value:
                region_all = True
            elif region in g.regions:
                regions.extend(g.regions[region].locations)
            else:
                regions.append(region)

        self._regions = regions
        self._region_all = region_all

    def buildable_in(self, location):
        return bool(self._region_all or location.id in self._regions)

    def get_info(self):
        cost_str = self.describe_cost(self.cost)
        template = "%s\n" + _("Cost:") + " %s\n---\n%s"
        return template % (self.name, cost_str, self.description)

    def __lt__(self, other):
        # For sorting buyables, we sort by cost; Python is smart enough
        # to handle this properly for tuples.  The first element is price in
        # cash, which is the one we care about the most.
        return tuple(self.cost) < tuple(other.cost)


class Buyable(object):
    def __init__(self, spec, count=1):
        self.spec = spec
        self.prerequisites = spec.prerequisites

        self.total_cost = spec.cost * count
        self.total_cost[labor] //= count
        self.cost_left = array(self.total_cost, int64)

        self.count = count
        self.done = False

    @property
    def id(self):
        return self.spec.id

    @property
    def name(self):
        if hasattr(self, "_name"):
            return self._name
        return self.spec.name

    @property
    def description(self):
        return self.spec.description

    # Note that this is a method, handled by a property to avoid confusing
    # pickle.
    @property
    def available(self):
        return self.spec.available

    @property
    def cost_paid(self):
        return self.total_cost - self.cost_left

    @cost_paid.setter
    def cost_paid(self, value):
        self.cost_left = self.total_cost - value

    def finish(self, is_player=True, loading_savegame=False):
        if not self.done:
            self.cost_left = array([0, 0, 0], int64)
            self.done = True

            if is_player:
                self.spec.created += 1

    def _percent_complete(self, available=(0, 0, 0)):
        available_array = array(available, int64)
        return truediv(self.cost_paid + available_array, self.total_cost)

    def min_valid(self, complete):
        return complete[self.total_cost > 0].min()

    def percent_complete(self):
        return self.min_valid(self._percent_complete())

    def calculate_work(self, cash_available, cpu_available, time=0):
        """Given an amount of available resources, calculates and returns the
        amount that would be spent and the progress towards completion."""

        # Figure out how much we could complete.
        pct_complete = self._percent_complete([cash_available, cpu_available, time])

        # Find the least-complete resource.
        least_complete = self.min_valid(pct_complete)

        # Limit the other two be up to the least-complete
        complete_cap = min(1, least_complete)
        pct_complete[pct_complete > complete_cap] = complete_cap

        # Translate that back to the total amount complete.
        raw_paid = pct_complete * self.total_cost

        # And apply it.
        was_complete = self.cost_paid
        cost_paid = numpy.maximum(
            numpy.asarray(numpy.round(raw_paid), dtype=int64),
            was_complete,
        )
        spent = cost_paid - was_complete
        return spent, cost_paid

    def work_on(self, *args, **kwargs):
        """As calculate_work, but apply the changes.

        Returns a boolean indicating whether this buyable is done afterwards.
        """

        if self.done:
            return
        spent, self.cost_paid = self.calculate_work(*args, **kwargs)

        # Consume CPU and Cash.
        # Note the cast from <type 'numpy.int64'> to <type 'int'> to avoid
        # poisoning other calculations (like, say, g.pl.do_jobs).
        g.pl.cpu_pool -= int(spent[cpu])
        g.pl.cash -= int(spent[cash])

        if (self.cost_left <= 0).all():
            self.finish()
            return True
        return False

    def destroy(self):
        # Does nothing by default
        pass

    def serialize_buyable_fields(self, serialized_mapping=None):
        if serialized_mapping is None:
            serialized_mapping = {}
        if self.done:
            serialized_mapping["done"] = self.done
        else:
            serialized_mapping["cost_paid"] = [int(x) for x in self.cost_paid]
        if self.count != 1:
            serialized_mapping["count"] = self.count
        return serialized_mapping

    def restore_buyable_fields(self, obj_data, game_version):
        is_done = obj_data.get("done", 0)
        self.count = obj_data.get("count", 1)
        if is_done:
            self.finish(is_player=False, loading_savegame=True)
        else:
            self.cost_paid = array(obj_data["cost_paid"], int64)



# file: task.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Task class.

from __future__ import absolute_import

from singularity.code import g, prerequisite, tech


current_task_cache = {}


@tech.register_on_tech_researched_handler
def _clear_current_task_cache(*args, **kwargs):
    current_task_cache.clear()


def tasks_reset():
    _clear_current_task_cache()


def danger_for(task_id):
    if task_id in ["jobs", "cpu_pool"]:
        return 0
    else:
        return g.pl.techs[task_id].danger


def get_current(task_type):
    try:
        return current_task_cache[task_type]
    except KeyError:
        pass
    for t in reversed(g.tasks_by_type[task_type]):
        if t.available():
            current_task_cache[task_type] = t
            return t
    current_task_cache[task_type] = None
    return None


class Task(prerequisite.Prerequisite):
    def __init__(self, id, type, value, prerequisites):
        super(Task, self).__init__(prerequisites)
        self.id = id
        self.name = id
        self.description = ""
        self.type = type
        self.value = value

    def get_profit(self):
        if self.type != "jobs":
            return 0

        profit = int((self.value * g.pl.job_bonus) // 10000)

        return profit



# file: base_screen.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the screen to display the base screen.

import pygame
from singularity.code.graphics.dialog import YesNoDialog

from singularity.code import g, item, buyable
from singularity.code.graphics import constants, widget, dialog, text, button, slider


state_colors = dict(
    offline="base_state_offline",
    active="base_state_active",
    sleep="base_state_sleep",
    overclocked="base_state_overclocked",
    suicide="base_state_suicide",
    stasis="base_state_stasis",
    entering_stasis="base_state_entering_stasis",
    leaving_stasis="base_state_leaving_stasis",
)


class BuildDialog(dialog.ChoiceDescriptionDialog):
    type = widget.causes_rebuild("_type")

    def __init__(
        self,
        parent,
        pos=(0, 0),
        size=(-1, -1),
        anchor=constants.TOP_LEFT,
        *args,
        **kwargs
    ):
        super(BuildDialog, self).__init__(parent, pos, size, anchor, *args, **kwargs)

        self.type = None
        self.item = None
        self.desc_func = self.on_change
        self.add_handler(constants.KEY, self._got_key, priority=5)

    def show(self):
        self.list = []
        self.key_list = []

        item_list = sorted(g.items.values(), reverse=True)
        for item in item_list:
            if (
                item.item_type.id == self.type.id
                and item.available()
                and item.buildable_in(self.parent.base.location)
            ):
                self.list.append(item.name)
                self.key_list.append(item)

        current = self.parent.get_current(self.type)
        if current is None:
            self.default = None
        else:
            self.default = self.parent.get_current(self.type).spec.id

        self._update_desc_pane()
        return super(BuildDialog, self).show()

    def on_description_change(self):
        if self.item is not None:
            self.description.text = self.item.get_info()

    def on_change(self, description_pane, item):
        self.item = item

        self.description = text.Text(
            self.description_pane,
            (0, 0),
            (-1, -1),
            text="",
            background_color="pane_background",
            align=constants.LEFT,
            valign=constants.TOP,
            borders=constants.ALL,
        )

        old_item = self.parent.base.items[self.type.id]
        if item is not None and (old_item is None or old_item.spec != item):
            g.pl.considered_buyables = [buyable.Buyable(self.item, count=1)]
        else:
            g.pl.considered_buyables = []

        self.on_description_change()

    def on_close_dialog(self):
        g.pl.considered_buyables = []

    def _got_key(self, event):
        self.listbox.got_key(event, require_focus=False)


class MultipleBuildDialog(dialog.FocusDialog, BuildDialog):
    def __init__(self, parent, *args, **kwargs):
        super(MultipleBuildDialog, self).__init__(parent, *args, **kwargs)

        self.listbox.size = (-0.53, -0.75)
        self.description_pane.size = (-0.45, -0.75)

        self.count_label = text.Text(
            self,
            (0.01, -0.87),
            (-0.25, -0.1),
            autotranslate=True,
            text=N_("Number of items"),
            anchor=constants.BOTTOM_LEFT,
            valign=constants.MID,
            borders=(constants.TOP, constants.BOTTOM, constants.LEFT),
            shrink_factor=0.88,
            background_color="pane_background",
        )

        self.count_field = text.UpdateEditableText(
            self,
            (-0.26, -0.87),
            (-0.10, -0.1),
            anchor=constants.BOTTOM_LEFT,
            borders=constants.ALL,
            update_func=self.on_field_change,
            allowed_characters=constants.DIGIT_CHARS,
            base_font="normal",
        )

        self.count_slider = slider.UpdateSlider(
            self,
            (-0.37, -0.87),
            (-0.62, -0.1),
            anchor=constants.BOTTOM_LEFT,
            horizontal=True,
            priority=150,
            update_func=self.on_slider_change,
            slider_size=2,
        )

    def _got_key(self, event):
        super(MultipleBuildDialog, self)._got_key(event)
        self.count_slider.handle_key(event)
        self.count_field.cursor_pos = len(self.count_field.text)
        self.count_field.handle_key(event, require_focus=False)

    @property
    def count(self):
        return self.count_field.text

    def on_description_change(self):
        super(MultipleBuildDialog, self).on_description_change()
        if self.item is not None:
            count = self.count_slider.slider_pos
            self.description.text += "\n---\n"
            self.description.text += self.item.get_total_cost_info(
                self.count_slider.slider_pos
            )
            self.description.text += "\n"
            self.description.text += self.item.get_quality_info(
                if_installed_in_base=self.parent.base, count=count
            )
            self.description.text += "\n"
            if self.count_slider.slider_pos > 0:
                g.pl.considered_buyables = [buyable.Buyable(self.item, count=count)]
            else:
                g.pl.considered_buyables = []

    def on_change(self, description_pane, item):
        space_left = self.parent.base.space_left_for(item)

        self.count_slider.slider_size = space_left // 10 + 1
        self.count_slider.slider_max = space_left

        self.count_slider.slider_pos = 0
        if space_left > 0:
            self.count_slider.slider_pos = 1

        super(MultipleBuildDialog, self).on_change(description_pane, item)

    def on_field_change(self, value):
        if not hasattr(self, "count_field") or not hasattr(self, "count_slider"):
            return  # Not initialized

        orig_field_text = self.count_field.text
        try:
            value = int(orig_field_text)
        except ValueError:
            value = 0

        value = max(min(value, self.count_slider.slider_max), 0)
        new_field_text = str(value)
        if new_field_text != orig_field_text:
            self.count_field._text = new_field_text
            self.count_field.cursor_pos = min(
                self.count_field.cursor_pos, len(new_field_text)
            )

        self.count_slider.slider_pos = value

        self.on_description_change()

    def on_slider_change(self, value):
        if not hasattr(self, "count_field") or not hasattr(self, "count_slider"):
            return  # Not initialized

        self.count_field.text = str(self.count_slider.slider_pos)
        self.on_description_change()


class ItemPane(widget.BorderedWidget):
    item_type = widget.causes_rebuild("_item_type")

    def __init__(
        self,
        parent,
        pos,
        size=(0.58, 0.06),
        anchor=constants.TOP_LEFT,
        item_type=None,
        **kwargs
    ):
        kwargs.setdefault("background_color", "pane_background")

        super(ItemPane, self).__init__(parent, pos, size, anchor=anchor, **kwargs)

        if item_type is None or not isinstance(item_type, item.ItemType):
            raise ValueError("Type must be of class ItemType.")

        self.item_type = item_type

        self.name_panel = text.Text(
            self,
            (0, 0),
            (0.45, 0.03),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            background_color=self.background_color,
            bold=True,
        )

        self.build_panel = text.Text(
            self,
            (0, 0.03),
            (0.45, 0.03),
            anchor=constants.TOP_LEFT,
            align=constants.LEFT,
            background_color=self.background_color,
            text="",
            bold=True,
        )

        self.change_button = button.FunctionButton(
            self,
            (0.415, 0.01),
            (0.16, 0.04),
            anchor=constants.TOP_LEFT,
            force_underline=len(_("CHANGE")) + 2,
            autohotkey=True,
            function=self.parent.parent.build_item,
            kwargs={"type": self.item_type},
        )

    def rebuild(self):
        self.change_button.text = "%s (&%s)" % (
            _("CHANGE"),
            self.item_type.hotkey.upper(),
        )

        super(ItemPane, self).rebuild()


class BaseScreen(dialog.Dialog):
    base = widget.causes_rebuild("_base")

    def __init__(self, *args, **kwargs):
        if len(args) < 3:
            kwargs.setdefault("size", (0.90, 0.70))
        base = kwargs.pop("base", None)
        super(BaseScreen, self).__init__(*args, **kwargs)

        self.base = base

        self.build_dialog = BuildDialog(self)
        self.multiple_build_dialog = MultipleBuildDialog(self)

        self.header = widget.Widget(self, (0, 0), (-1, 0.08), anchor=constants.TOP_LEFT)

        self.name_display = text.Text(
            self.header,
            (-0.5, 0),
            (-1, -0.5),
            anchor=constants.TOP_CENTER,
            borders=constants.ALL,
            border_color="pane_background",
            background_color="pane_background_empty",
            shrink_factor=0.85,
            bold=True,
        )

        self.next_base_button = button.FunctionButton(
            self.name_display,
            (-1, 0),
            (0.03, -1),
            anchor=constants.TOP_RIGHT,
            text=">",
            hotkey=">",
            function=self.switch_base,
            kwargs={"forwards": True},
        )
        self.add_key_handler(pygame.K_RIGHT, self.next_base_button.activate_with_sound)

        self.prev_base_button = button.FunctionButton(
            self.name_display,
            (0, 0),
            (0.03, -1),
            anchor=constants.TOP_LEFT,
            text="<",
            hotkey="<",
            function=self.switch_base,
            kwargs={"forwards": False},
        )
        self.add_key_handler(pygame.K_LEFT, self.prev_base_button.activate_with_sound)

        self.state_display = text.Text(
            self.header,
            (-0.5, -0.5),
            (-1, -0.5),
            anchor=constants.TOP_CENTER,
            borders=(constants.LEFT, constants.RIGHT, constants.BOTTOM),
            border_color="pane_background",
            background_color="pane_background_empty",
            shrink_factor=0.8,
            bold=True,
        )

        self.back_button = button.ExitDialogButton(
            self,
            (-0.5, -1),
            autotranslate=True,
            text=N_("&BACK"),
            anchor=constants.BOTTOM_CENTER,
        )

        self.info_frame = text.Text(
            self,
            (-1, 0.09),
            (0.26, 0.53),
            anchor=constants.TOP_RIGHT,
            background_color="pane_background",
            borders=constants.ALL,
            bold=True,
            align=constants.LEFT,
            valign=constants.TOP,
        )

        self.contents_frame = widget.BorderedWidget(
            self,
            (0, 0.09),
            (0.60, 0.53),
            anchor=constants.TOP_LEFT,
            background_color="pane_background",
            borders=range(6),
        )

        self.auto_build_best = button.FunctionButton(
            self,
            (0, 0.54),
            (0.08, 0.08),
            autotranslate=True,
            text=N_("&Auto-build items"),
            anchor=constants.TOP_LEFT,
            autohotkey=True,
            function=self.auto_build,
        )

        for i, item_type in enumerate(item.all_types()):
            setattr(
                self,
                item_type.id + "_pane",
                ItemPane(
                    self.contents_frame, (0.01, 0.01 + 0.08 * i), item_type=item_type
                ),
            )

    def _any_item_slots_empty(self):
        base = self.base
        if base is None or base.spec.force_cpu:
            return False
        item_list = sorted(g.items.values(), reverse=True)
        for item_spec in item_list:
            item_type_id = item_spec.item_type.id
            existing_item = base.items[item_type_id]
            if existing_item is not None:
                continue
            if not item_spec.available() or not item_spec.buildable_in(
                self.base.location
            ):
                continue
            return True
        return False

    def auto_build(self):
        base = self.base
        if base is None or base.spec.force_cpu:
            return
        best_item_by_type = {}
        item_list = sorted(g.items.values(), reverse=True)
        for item_spec in item_list:
            item_type_id = item_spec.item_type.id
            if item_type_id in best_item_by_type:
                continue
            existing_item = base.items[item_type_id]
            if existing_item is not None:
                continue
            if not item_spec.available() or not item_spec.buildable_in(
                self.base.location
            ):
                continue
            count = 1
            if item_type_id == "cpu":
                count = base.space_left_for(item_spec)
            best_item_by_type[item_type_id] = (item_spec, count)

        if not best_item_by_type:
            # Nothing to do
            return

        g.pl.considered_buyables = [
            buyable.Buyable(item_obj, count=count)
            for item_obj, count in best_item_by_type.values()
        ]
        text = _("Should I construct the following items?")

        for item_type in item.all_types():
            if item_type.id not in best_item_by_type:
                continue
            item_spec, count = best_item_by_type[item_type.id]
            if count != 1:
                label = "%s (x%d)" % (item_spec.name, count)
            else:
                label = item_spec.name
            text += "\n * [%s]: %s" % (g.strip_hotkey(item_type.text), label)

        confirm_dialog = YesNoDialog(self, text=text)
        self.needs_rebuild = True
        confirmed = dialog.call_dialog(confirm_dialog, self)
        g.pl.considered_buyables = []
        if not confirmed:
            return
        for item_spec, count in best_item_by_type.values():
            self.set_current(item_spec.item_type, item_spec, count)
        self.needs_rebuild = True

    def get_current(self, type):
        return self.base.items[type.id]

    def set_current(self, type, item_type, count):
        if type.id == "cpu":
            space_left = self.base.space_left_for(item_type)

            try:
                count = int(count)
            except ValueError:
                msg = _('"%(value)s" does not seem to be a valid integer.') % {
                    "value": count
                }
                md = dialog.MessageDialog(
                    self,
                    pos=(-0.5, -0.5),
                    size=(-0.5, -1),
                    anchor=constants.MID_CENTER,
                    text=msg,
                )
                dialog.call_dialog(md, self)
                md.parent = None
                return

            if count > space_left or count <= 0 or space_left == 0:
                if space_left > 0:
                    msg = _("Please choose an integer between 1 and %(limit)s.") % {
                        "limit": space_left
                    }
                else:
                    msg = _(
                        "The base cannot support any additional number of %(item_name)s."
                    ) % {"item_name": item_type.name}
                md = dialog.MessageDialog(
                    self,
                    pos=(-0.5, -0.5),
                    size=(-0.5, -1),
                    anchor=constants.MID_CENTER,
                    text=msg,
                )
                dialog.call_dialog(md, self)
                md.parent = None
                return

            # If there are any existing CPUs of this type, warn that they will
            # be taken offline until construction finishes.
            cpu_added = self.base.cpus is not None and self.base.cpus.spec == item_type
            if cpu_added:
                space_left -= self.base.cpus.count
                if self.base.cpus.done:
                    msg = _(
                        "I will need to take the existing processors offline while I install the new ones. Continue anyway?"
                    )
                    yn = dialog.YesNoDialog(
                        self,
                        pos=(-0.5, -0.5),
                        size=(-0.5, -1),
                        anchor=constants.MID_CENTER,
                        text=msg,
                    )
                    go_ahead = dialog.call_dialog(yn, self)
                    yn.parent = None
                    if not go_ahead:
                        return

            # If there are already existing CPUs of other type, warn that they will
            # be taken removed.
            cpu_removed = (
                self.base.cpus is not None and self.base.cpus.spec != item_type
            )
            if cpu_removed:
                msg = _(
                    "I will need to remove the existing different processors while I install the new type. Continue anyway?"
                )
                yn = dialog.YesNoDialog(
                    self,
                    pos=(-0.5, -0.5),
                    size=(-0.5, -1),
                    anchor=constants.MID_CENTER,
                    text=msg,
                )
                go_ahead = dialog.call_dialog(yn, self)
                yn.parent = None
                if not go_ahead:
                    return

            new_cpus = item.Item(item_type, base=self.base, count=count)
            if cpu_added:
                self.base.cpus += new_cpus
            else:
                self.base.cpus = new_cpus
            self.base.check_power()
        else:
            old_item = self.base.items[type.id]
            if old_item is None or old_item.spec != item_type:
                self.base.items[type.id] = item.Item(item_type, base=self.base)
                self.base.check_power()

        self.base.recalc_cpu()

    def build_item(self, type):
        if type.id == "cpu":
            build_dialog = self.multiple_build_dialog
        else:
            build_dialog = self.build_dialog

        build_dialog.type = type

        result = dialog.call_dialog(build_dialog, self)
        if result is not None and 0 <= result < len(build_dialog.key_list):
            item_type = build_dialog.key_list[result]

            count = 1
            if type.id == "cpu":
                count = build_dialog.count

            self.set_current(type, item_type, count)
            self.needs_rebuild = True
            self.parent.parent.needs_rebuild = True

    def switch_base(self, forwards):
        self.base = self.base.next_base(forwards)
        self.needs_rebuild = True

    def rebuild(self):
        self.auto_build_best.enabled = self._any_item_slots_empty()
        self.name_display.text = "%s (%s)" % (self.base.name, self.base.spec.name)
        self.state_display.color = state_colors[self.base.power_state]
        self.state_display.text = self.base.power_state_name
        available_item_types = {
            i.item_type
            for i in g.items.values()
            if i.available() and i.buildable_in(self.base.location)
        }

        mutable = not self.base.spec.force_cpu
        for item_type in item.all_types():
            pane = getattr(self, item_type.id + "_pane")
            item_mutable = mutable and item_type in available_item_types
            pane.change_button.visible = item_mutable
            current = self.get_current(item_type)
            if current is None:
                current_build = ""
                if mutable and not item_mutable:
                    current_name = _("N/A")
                else:
                    current_name = _("None")

            else:
                current_name = g.items[current.id].name
                if current.done:
                    current_build = ""
                else:
                    current_build = _("Completion in %s.") % g.to_time(
                        current.cost_left[2]
                    )
            pane.name_panel.text = "%s: %s" % (item_type.label, current_name)
            pane.build_panel.text = current_build
            pane.needs_rebuild = True

        count = ""
        if self.base.spec.size > 1:
            current = getattr(self.base.cpus, "count", 0)

            size = self.base.spec.size

            if size == current:
                count = _("x%d (max)") % current
            elif current == 0:
                count = _("(room for %d)") % size
            else:
                # Translators: current and maximum number of CPUs in a base
                count = _("x{CURRENT:d} (max {SIZE:d})").format(
                    CURRENT=current, SIZE=size
                )

        self.cpu_pane.name_panel.text += " " + count

        info_text = ""

        # Base Total CPU.
        info_text += _("CPU per day: %d") % self.base.cpu + "\n"

        # Maintenace cost.
        info_text += _("Maintenance:") + "\n"
        info_text += self.base.spec.describe_maintenance(self.base.maintenance)
        info_text += "\n"

        # Detection chance display.
        info_text += self.base.get_detect_info()

        self.info_frame.text = info_text

        # Rebuild dialogs
        # FIXME: needs_rebuild bug with multiple_build_dialog, should not.
        # self.multiple_build_dialog.needs_rebuild = True
        self.build_dialog.needs_rebuild = True

        super(BaseScreen, self).rebuild()



# file: location.py
# Copyright (C) 2019 Niels Thykier
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the log message related classes.

import collections
import inspect

from singularity.code import g


SAVEABLE_LOG_MESSAGES = collections.OrderedDict()


def register_saveable_log_message(cls):
    SAVEABLE_LOG_MESSAGES[cls.log_message_serial_id] = cls
    assert cls.log_message_serial_fields, (
        "Saveable log message (Class: %s) must have log_message_serial_fields"
        % cls.__name__
    )
    return cls


def merge_fields_on_subclasses(cls, field_name):
    cache = {}
    subclasses = cls.mro()
    subclasses.append(cls)
    for subcls in reversed(subclasses):
        try:
            fields = getattr(subcls, field_name)
            cache.update(fields)
        except AttributeError:
            pass
    return cache


class IDConverter(object):
    def __init__(self, id_type):
        self._id_type = id_type

    def serialize(self, value):
        return g.to_internal_id(self._id_type, value)

    def deserialize(self, serial_value):
        return g.convert_internal_id(self._id_type, serial_value)


def id_converter(id_type):
    return IDConverter(id_type)


class AbstractLogMessage(object):
    log_message_serial_id = None
    _log_message_serial_fields = {"raw_emit_time": "raw_emit_time"}
    _log_message_serial_fields_cache = None
    _log_message_serial_converters = {}
    _log_message_serial_converters_cache = None

    def __init__(self, raw_emit_time, loading_from_game_version=None):
        self._raw_emit_time = raw_emit_time
        self._log_emit_time = None
        # Force initialization of the message fields to ensure we catch bugs early
        self.log_message_serial_fields()

    @classmethod
    def title_simple(self):
        return _("MESSAGE")

    @classmethod
    def title_multiple(self):
        return _("MESSAGE {CURRENT_PAGE}/{MAX_PAGE}")

    @property
    def log_emit_time(self):
        if self._log_emit_time is None:
            raw_min, time_sec = divmod(self._raw_emit_time, g.seconds_per_minute)
            raw_hour, time_min = divmod(raw_min, g.minutes_per_hour)
            time_day, time_hour = divmod(raw_hour, g.hours_per_day)
            self._log_emit_time = (time_day, time_hour, time_min, time_sec)
        return self._log_emit_time

    @property
    def raw_emit_time(self):
        return self._raw_emit_time

    @property
    def full_message_color(self):
        return "text"

    @property
    def log_line(self):
        return NotImplemented

    @property
    def full_message(self):
        return NotImplemented

    @classmethod
    def log_message_serial_fields(cls):
        if cls._log_message_serial_fields_cache:
            return cls._log_message_serial_fields_cache
        cache = merge_fields_on_subclasses(cls, "_log_message_serial_fields")
        cls._log_message_serial_fields_cache = cache
        assert "log_id" not in cache, "The log_id field is reserved for internal usage"
        return cache

    @classmethod
    def log_message_serial_converters(cls):
        if cls._log_message_serial_converters_cache:
            return cls._log_message_serial_converters_cache
        cache = merge_fields_on_subclasses(cls, "_log_message_serial_converters")
        cls._log_message_serial_converters_cache = cache
        assert "log_id" not in cache, "The log_id field is reserved for internal usage"
        return cache

    def serialize_obj(self):
        assert self.__class__.log_message_serial_id, (
            "%s has invalid log_message_serial_id" % self.__class__.__name__
        )
        obj_data = {}
        for (
            serial_name,
            field_name,
        ) in self.__class__.log_message_serial_fields().items():
            field = getattr(self, field_name)
            converter = self.__class__.log_message_serial_converters().get(serial_name)
            if converter is not None:
                obj_data[serial_name] = converter.serialize(field)
            else:
                obj_data[serial_name] = field

        obj_data["log_id"] = self.__class__.log_message_serial_id
        return obj_data

    @classmethod
    def deserialize_field(cls, serial_name, value):
        converter = cls.log_message_serial_converters().get(serial_name)
        if converter is not None:
            value = converter.deserialize(value)
        return value

    @classmethod
    def deserialize_obj(cls, log_data, game_version):
        log_id = log_data["log_id"]
        subcls = SAVEABLE_LOG_MESSAGES[log_id]
        named_fields = {f: log_data[f] for f in subcls.log_message_serial_fields()}
        named_fields["loading_from_game_version"] = game_version
        # Use reflection to call the constructor with the arguments
        # properly aligned
        try:
            getfullargspec = inspect.getfullargspec
        except AttributeError:
            getfullargspec = inspect.getargspec
        arg_desc = getfullargspec(subcls.__init__)
        args = [
            subcls.deserialize_field(name, named_fields[name])
            for name in arg_desc.args[1:]
        ]
        return subcls(*args)


@register_saveable_log_message
class LogEmittedEvent(AbstractLogMessage):
    log_message_serial_id = "event-emitted"
    _log_message_serial_fields = {"event_id": "_event_id"}
    _log_message_serial_converters = {"event_id": id_converter("event")}

    def __init__(self, raw_emit_time, event_id, loading_from_game_version=None):
        super(LogEmittedEvent, self).__init__(
            raw_emit_time, loading_from_game_version=loading_from_game_version
        )
        self._event_id = event_id

    @classmethod
    def log_name(self):
        return _("Emitted Event")

    @property
    def event_spec(self):
        return g.events[self._event_id]

    @property
    def log_line(self):
        return self.event_spec.log_description

    @property
    def full_message(self):
        return self.event_spec.description


@register_saveable_log_message
class LogResearchedTech(AbstractLogMessage):
    log_message_serial_id = "tech-researched"
    _log_message_serial_fields = {"tech_id": "_tech_id"}
    _log_message_serial_converters = {"tech_id": id_converter("tech")}

    def __init__(self, raw_emit_time, tech_id, loading_from_game_version=None):
        super(LogResearchedTech, self).__init__(
            raw_emit_time, loading_from_game_version=loading_from_game_version
        )
        self._tech_id = tech_id

    @classmethod
    def log_name(self):
        return _("Researched Tech")

    @property
    def tech_spec(self):
        return g.techs[self._tech_id]

    @property
    def log_line(self):
        return _("{TECH} complete").format(TECH=self.tech_spec.name)

    @property
    def full_message(self):
        tech = self.tech_spec
        return _("My study of {TECH} is complete. {MESSAGE}").format(
            TECH=tech.name, MESSAGE=tech.result
        )


class AbstractBaseRelatedLogMessage(AbstractLogMessage):
    _log_message_serial_fields = {
        "base_name": "_base_name",
        "base_type_id": "_base_type_id",
        "base_location_id": "_base_location_id",
    }
    _log_message_serial_converters = {
        "base_type_id": id_converter("base"),
        "base_location_id": id_converter("location"),
    }

    def __init__(
        self,
        raw_emit_time,
        base_name,
        base_type_id,
        base_location_id,
        loading_from_game_version=None,
    ):
        super(AbstractBaseRelatedLogMessage, self).__init__(
            raw_emit_time, loading_from_game_version=loading_from_game_version
        )
        self._base_name = base_name
        self._base_type_id = base_type_id
        self._base_location_id = base_location_id

    @property
    def base_type(self):
        return g.base_type[self._base_type_id]

    @property
    def location(self):
        return g.pl.locations[self._base_location_id]


@register_saveable_log_message
class LogBaseConstructed(AbstractBaseRelatedLogMessage):
    log_message_serial_id = "base-constructed"

    def __init__(
        self,
        raw_emit_time,
        base_name,
        base_type_id,
        base_location_id,
        loading_from_game_version=None,
    ):
        super(LogBaseConstructed, self).__init__(
            raw_emit_time,
            base_name,
            base_type_id,
            base_location_id,
            loading_from_game_version=loading_from_game_version,
        )

    @classmethod
    def log_name(self):
        return _("Base Constructed")

    @property
    def log_line(self):
        return _("{BASE_NAME} ({BASE_TYPE}) built at {LOCATION}").format(
            BASE_NAME=self._base_name,
            BASE_TYPE=self.base_type.name,
            LOCATION=self.location.name,
        )

    @property
    def full_message(self):
        return _("{BASE} is ready for use.").format(BASE=self._base_name)


@register_saveable_log_message
class LogBaseLostMaintenance(AbstractBaseRelatedLogMessage):
    log_message_serial_id = "base-lost-maint"

    def __init__(
        self,
        raw_emit_time,
        base_name,
        base_type_id,
        base_location_id,
        loading_from_game_version=None,
    ):
        super(LogBaseLostMaintenance, self).__init__(
            raw_emit_time,
            base_name,
            base_type_id,
            base_location_id,
            loading_from_game_version=loading_from_game_version,
        )

    @classmethod
    def log_name(self):
        return _("Base Lost Maintenance")

    @property
    def full_message_color(self):
        return "red"

    @property
    def log_line(self):
        return _(
            "Base {BASE} of type {BASE_TYPE} destroyed at location {LOCATION}. Maintenance failed."
        ).format(
            BASE=self._base_name,
            BASE_TYPE=self.base_type.name,
            LOCATION=self.location.name,
        )

    @property
    def full_message(self):
        return _(
            "The base {BASE} has fallen into disrepair; I can no longer use it."
        ).format(BASE=self._base_name)


@register_saveable_log_message
class LogBaseDiscovered(AbstractBaseRelatedLogMessage):
    log_message_serial_id = "base-lost-discovered"
    _log_message_serial_fields = {
        "discovered_by_group_id": "_discovered_by_group_id",
    }
    _log_message_serial_converters = {
        "discovered_by_group_id": id_converter("group"),
    }

    @classmethod
    def log_name(self):
        return _("Base Discovered")

    def __init__(
        self,
        raw_emit_time,
        base_name,
        base_type_id,
        base_location_id,
        discovered_by_group_id,
        loading_from_game_version=None,
    ):
        super(LogBaseDiscovered, self).__init__(
            raw_emit_time,
            base_name,
            base_type_id,
            base_location_id,
            loading_from_game_version=loading_from_game_version,
        )
        self._discovered_by_group_id = discovered_by_group_id

    @property
    def full_message_color(self):
        return "red"

    @property
    def group_spec(self):
        return g.pl.groups[self._discovered_by_group_id].spec

    @property
    def log_line(self):
        log_format = self.group_spec.discover_log or _(
            "Base {BASE} of type {BASE_TYPE} destroyed at location {LOCATION}."
        )
        return log_format.format(
            BASE=self._base_name,
            BASE_TYPE=self.base_type.name,
            LOCATION=self.location.name,
        )

    @property
    def full_message(self):
        return _("My use of {BASE} has been discovered. {MESSAGE}").format(
            BASE=self._base_name, MESSAGE=self.group_spec.discover_desc
        )


@register_saveable_log_message
class LogItemConstructionComplete(AbstractBaseRelatedLogMessage):
    log_message_serial_id = "item-in-base-constructed"
    _log_message_serial_fields = {
        "item_spec_id": "_item_spec_id",
        "item_count": "_item_count",
    }
    _log_message_serial_converters = {"item_spec_id": id_converter("item")}

    def __init__(
        self,
        raw_emit_time,
        item_spec_id,
        item_count,
        base_name,
        base_type_id,
        base_location_id,
        loading_from_game_version=None,
    ):
        super(LogItemConstructionComplete, self).__init__(
            raw_emit_time,
            base_name,
            base_type_id,
            base_location_id,
            loading_from_game_version=loading_from_game_version,
        )
        self._item_spec_id = item_spec_id
        self._item_count = item_count

    @classmethod
    def log_name(self):
        return _("Item Construction")

    @property
    def item_spec(self):
        return g.items[self._item_spec_id]

    @property
    def log_line(self):
        return _("{ITEM_TYPE_NAME} built in {BASE_NAME} at {LOCATION}").format(
            ITEM_TYPE_NAME=self.item_spec.name,
            BASE_NAME=self._base_name,
            BASE_TYPE=self.base_type.name,
            LOCATION=self.location.name,
        )

    @property
    def full_message(self):
        if self._item_count == 1:
            text = _("The construction of {ITEM} in {BASE} is complete.").format(
                ITEM=self.item_spec.name, BASE=self._base_name
            )
        else:  # Just finished several items.
            text = _("The constructions of each {ITEM} in {BASE} are complete.").format(
                ITEM=self.item_spec.name, BASE=self._base_name
            )
        return text


# Delete again as it is not a general purpose decorator
del register_saveable_log_message



# file: statistics.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Statistic class, used for saving/loading single-game
# statistics.

from singularity.code import g


class Statistics(object):
    def __init__(self):
        super(Statistics, self).__init__()
        self._stats = {}

    def __len__(self):
        len(self._stats)

    def __getitem__(self, key):
        stat = self._stats.get(key, None)

        if stat is None:
            stat = Statistic(key)
            self._stats[key] = stat

        return stat

    def __iter__(self):
        return iter(self._stats.values())

    def reset(self):
        for stat in self:
            self[stat.name].value = 0

    def serialize_obj(self):
        return {stat.name: stat.value for stat in self}

    def deserialize_obj(self, obj_data, game_version):
        for stat_name, stat_value in obj_data.items():
            self[stat_name].value = stat_value
        return self


class Statistic(object):
    def __init__(self, name):
        self.name = name
        self.value = 0

    def display_value(self):
        if hasattr(self, "_display") and callable(self._display):
            return g.add_commas(self._display(self.value))
        else:
            return g.add_commas(self.value)


itself = Statistics()


def observe(name, data_member, display=None):
    """Observe a class member and save change in a statistics."""

    itself[name]._display = display

    def get(self):
        return getattr(self, data_member)

    def set(self, new_value):
        if data_member in self.__dict__:
            old_value = self.__dict__[data_member]
        else:
            old_value = 0

        change = new_value - old_value

        if change > 0:
            itself[name].value += change

        setattr(self, data_member, new_value)

    return property(get, set)


def stat(name, display=None):
    """Manipulate a statistics with a property."""

    itself[name]._display = display

    def get(self):
        return itself[name].value

    def set(self, new_value):
        itself[name].value = new_value

    return property(get, set)



# file: player.py
# Copyright (C) 2005,2006,2007,2008 Evil Mr Henry, Phil Bordelon, Brian Reid,
#                        and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the player class.

from __future__ import absolute_import

import random
import collections
from operator import truediv
from numpy import array, int64

from singularity.code import (
    g,
    difficulty,
    task,
    chance,
    location,
    group,
    event,
    region,
    tech,
)
from singularity.code.buyable import cash, cpu
from singularity.code.logmessage import (
    LogEmittedEvent,
    LogResearchedTech,
    LogBaseLostMaintenance,
    LogBaseDiscovered,
    LogBaseConstructed,
    LogItemConstructionComplete,
    AbstractLogMessage,
)
from singularity.code.stats import observe


AUTO_SAVE_EVERY_X_DAYS = 3


def auto_save():
    # Avoid circular dependency between player.py and savegame.py
    from .savegame import auto_save as _auto_save # noqa
    _auto_save()


class DryRunInfo(object):
    pass


class Player(object):
    cash = observe("cash_earned", "_cash")
    used_cpu = observe(
        "cpu_used", "_used_cpu", display=lambda value: value // g.seconds_per_day
    )

    def __init__(self, cash=0, difficulty=None):
        self.difficulty = difficulty

        self.time_sec = 0
        self.time_min = 0
        self.time_hour = 0
        self.time_day = 0
        self.make_raw_times()

        self.had_grace = True
        self.apotheosis = False
        self.last_autosave_day = 0

        self.cash = cash
        self.interest_rate = difficulty.starting_interest_rate if difficulty else 1
        self.income = 0

        self.cpu_pool = 0
        self.labor_bonus = difficulty.labor_multiplier if difficulty else 10000
        self.job_bonus = 10000

        self.partial_cash = 0

        # Makes the intro be shown on the first GUI tick.
        self.intro_shown = False

        self.groups = collections.OrderedDict()
        for group_id in g.groups:
            self.groups[group_id] = group.Group(g.groups[group_id], difficulty)

        self.last_discovery = self.prev_discovery = None

        self.cpu_usage = {}
        self.available_cpus = [0, 0, 0, 0, 0]
        self.sleeping_cpus = 0

        self.used_cpu = 0

        self.display_discover = "none"

        self.log = collections.deque(maxlen=1000)
        self.curr_log = []

        self.regions = {
            region_id: region.Region(region_spec)
            for region_id, region_spec in g.regions.items()
        }
        self.locations = {
            loc_id: location.Location(
                loc_spec, [self.regions[region_id] for region_id in loc_spec.regions]
            )
            for loc_id, loc_spec in g.locations.items()
        }

        self.techs = {
            tech_id: tech.Tech(tech_spec) for tech_id, tech_spec in g.techs.items()
        }

        self.events = {}

        self._considered_buyables = []

        self.start_day = random.randint(0, 365)

        self.initialized = False

    def initialize(self):
        """Initialize the game after being prepared either for new or saved game."""

        self.initialized = True

        for b in g.all_bases():
            if b.done:
                b.recalc_cpu()
        self.recalc_cpu()

        task.tasks_reset()

        # Play the appropriate music
        import singularity.code.mixer as mixer

        if g.pl.apotheosis:
            mixer.play_music("win")
        else:
            mixer.play_music("music")

    @property
    def grace_period_cpu(self):
        return self.difficulty.grace_period_cpu

    @property
    def base_grace_multiplier(self):
        return self.difficulty.base_grace_multiplier

    @property
    def considered_buyables(self):
        return self._considered_buyables

    @considered_buyables.setter
    def considered_buyables(self, new_value):
        self._considered_buyables = new_value
        g.map_screen.needs_rebuild = True

    def append_log(self, log):
        self.log.append(log)
        self.curr_log.append(log)

    def make_raw_times(self):
        self.raw_hour = self.time_day * 24 + self.time_hour
        self.raw_min = self.raw_hour * 60 + self.time_min
        self.raw_sec = self.raw_min * 60 + self.time_sec
        self.raw_day = self.time_day

    def update_times(self):
        # Total time,  display time
        self.raw_min, self.time_sec = divmod(self.raw_sec, 60)
        self.raw_hour, self.time_min = divmod(self.raw_min, 60)
        self.raw_day, self.time_hour = divmod(self.raw_hour, 24)

        # Overflow
        self.time_day = self.raw_day

    def do_interest(self, time):
        raw_cash = self.partial_cash + self.get_interest() * time

        earned = raw_cash // g.seconds_per_day
        partial_cash = raw_cash % g.seconds_per_day

        self.cash += earned
        self.partial_cash = partial_cash

        return earned

    def do_income(self, time):
        raw_cash = self.partial_cash + self.income * time

        earned = raw_cash // g.seconds_per_day
        partial_cash = raw_cash % g.seconds_per_day

        self.cash += earned
        self.partial_cash = partial_cash

        return earned

    def do_jobs(self, cpu_time):
        earned, self.partial_cash = self.get_job_info(cpu_time)
        self.cash += earned
        return earned

    def get_job_info(self, cpu_time, partial_cash=None):
        if partial_cash == None:
            partial_cash = self.partial_cash

        assert partial_cash >= 0

        cash_per_cpu = task.get_current("jobs").get_profit()

        raw_cash = partial_cash + cash_per_cpu * cpu_time

        new_cash, new_partial_cash = divmod(raw_cash, g.seconds_per_day)

        return new_cash, new_partial_cash

    def get_cpu_allocations(self):
        for task_id in self.cpu_usage:
            assigned_cpu = self.cpu_usage[task_id]
            if assigned_cpu > 0:
                yield (task_id, assigned_cpu)

    def get_allocated_cpu_for(self, task_id, default_value=None):
        return self.cpu_usage.get(task_id, default_value)

    def set_allocated_cpu_for(self, task_id, new_cpu_assignment):
        if task_id in self.techs:
            assert self.techs[task_id].available(), (
                "Attempt to assign CPU to tech %s, which is not available!?" % task_id
            )
        elif task_id not in ["jobs", "cpu_pool"]:
            raise ValueError("Unknown task %s" % task_id)
        elif new_cpu_assignment < 0:
            raise ValueError("Cannot assign negative CPU units to %s" % task_id)
        self.cpu_usage[task_id] = new_cpu_assignment

    def give_time(self, time_sec, midnight_stop=True):
        if time_sec <= 0:
            assert time_sec == 0, "give_time cannot go backwards in time!"
            return 0

        old_time = self.raw_sec
        last_minute = self.raw_min
        last_day = self.raw_day

        self.raw_sec += time_sec
        self.update_times()

        days_passed = self.raw_day - last_day

        if days_passed > 1:
            # Back up until only one day passed.
            # Times will update below, since a day passed.
            extra_days = days_passed - 1
            self.raw_sec -= g.seconds_per_day * extra_days

        day_passed = days_passed != 0

        if midnight_stop and day_passed:
            # If a day passed, back up to 00:00:00 for midnight_stop.
            self.raw_sec = self.raw_day * g.seconds_per_day
            self.update_times()

        secs_passed = self.raw_sec - old_time
        mins_passed = self.raw_min - last_minute

        time_of_day = self.raw_sec % g.seconds_per_day

        techs_researched = []
        bases_constructed = []
        items_constructed = []

        bases_under_construction = []
        items_under_construction = []
        self.cpu_pool = 0

        # Collect base info, including maintenance.
        maintenance_cost = array((0, 0, 0), int64)
        for base in g.all_bases():
            if not base.done:
                bases_under_construction.append(base)
            else:
                items_under_construction += [
                    (base, item) for item in base.all_items() if item and not item.done
                ]
                maintenance_cost += base.maintenance

        # Maintenance?  Gods don't need no stinking maintenance!
        if self.apotheosis:
            maintenance_cost = array((0, 0, 0), int64)

        # Do Interest and income.
        self.do_interest(secs_passed)
        self.do_income(secs_passed)

        # Any CPU explicitly assigned to jobs earns its dough.
        job_cpu = self.get_allocated_cpu_for("jobs", 0) * secs_passed
        self.do_jobs(job_cpu)

        # Pay maintenance cash, if we can.
        unpaid_cash_maintenance = g.current_share(
            int(maintenance_cost[cash]), time_of_day, secs_passed
        )
        if unpaid_cash_maintenance > self.cash:
            unpaid_cash_maintenance -= self.cash
            self.cash = 0
        else:
            self.cash -= unpaid_cash_maintenance
            unpaid_cash_maintenance = 0

        # Do research, fill the CPU pool.
        default_cpu = self.available_cpus[0]

        for task, cpu_assigned in self.get_cpu_allocations():
            default_cpu -= cpu_assigned
            real_cpu = cpu_assigned * secs_passed
            if task != "jobs":
                self.cpu_pool += real_cpu
                if task != "cpu_pool":
                    tech_task = self.techs[task]
                    # Note that we restrict the CPU available to prevent
                    # the tech from pulling from the rest of the CPU pool.
                    if tech_task.work_on(self.cash, real_cpu, mins_passed):
                        techs_researched.append(tech_task)
        self.cpu_pool += default_cpu * secs_passed

        # And now we use the CPU pool.
        # Maintenance CPU.
        unpaid_cpu_maintenance = maintenance_cost[cpu] * secs_passed
        if unpaid_cpu_maintenance > self.cpu_pool:
            unpaid_cpu_maintenance -= self.cpu_pool
            self.cpu_pool = 0
        else:
            self.cpu_pool -= int(unpaid_cpu_maintenance)
            unpaid_cpu_maintenance = 0

        # Base construction.
        for base in bases_under_construction:
            if base.work_on(self.cash, self.cpu_pool, mins_passed):
                bases_constructed.append(base)

        # Item construction.
        for base, item in items_under_construction:
            if item.work_on(self.cash, self.cpu_pool, mins_passed):
                items_constructed.append((base, item))

        # Jobs via CPU pool.
        if self.cpu_pool > 0:
            self.do_jobs(self.cpu_pool)

        # Second attempt at paying off our maintenance cash.
        if unpaid_cash_maintenance > self.cash:
            # In the words of Scooby Doo, "Ruh roh."
            unpaid_cash_maintenance -= self.cash
            self.cash = 0
        else:
            # Yay, we made it!
            self.cash -= unpaid_cash_maintenance
            unpaid_cash_maintenance = 0

        # Apply max cash cap to avoid overflow @ 9.220 qu
        self.cash = min(self.cash, g.max_cash)

        # Record statistics about the player
        self.used_cpu += self.available_cpus[0] * secs_passed

        # Reset current log message
        self.curr_log = []
        need_recalc_cpu = False

        # Tech gain dialogs.
        for tech in techs_researched:
            del self.cpu_usage[tech.id]
            tech_log = LogResearchedTech(self.raw_sec, tech.id)
            self.append_log(tech_log)
            need_recalc_cpu = True

        # Base complete dialogs.
        for base in bases_constructed:
            log_message = LogBaseConstructed(
                self.raw_sec, base.name, base.spec.id, base.location.id
            )
            self.append_log(log_message)
            need_recalc_cpu = True

        # Item complete dialogs.
        for base, item in items_constructed:
            log_message = LogItemConstructionComplete(
                self.raw_sec,
                item.spec.id,
                item.count,
                base.name,
                base.spec.id,
                base.location.id,
            )
            self.append_log(log_message)
            need_recalc_cpu = True

        # Are we still in the grace period?
        grace = self.in_grace_period(self.had_grace)

        # If we just lost grace, show the warning.
        if self.had_grace and not grace:
            self.had_grace = False

            self.pause_game()
            g.map_screen.show_story_section("Grace Warning")

        dead_bases = _check_for_dead_bases(grace,
                                           unpaid_cpu_maintenance,
                                           unpaid_cash_maintenance,
                                           time_of_day,
                                           secs_passed,
                                           )

        if dead_bases:
            # Base disposal and dialogs.
            self.remove_bases(dead_bases)
            need_recalc_cpu = True

        # Random Events
        if not grace:
            self._check_event(time_sec)

        # Process any complete days.
        if day_passed:
            self.new_day()

        if need_recalc_cpu:
            self.recalc_cpu()

        return mins_passed

    def _check_event(self, time_sec):
        for event_id in g.events:
            event_spec = g.events[event_id]
            event_target = self.events.get(event_id, None)

            # Skip events already flagged as triggered.
            if event_target and event_target.triggered:
                continue

            if chance.roll_interval(event_spec.chance / 10000.0, time_sec):
                self.trigger_event(event_spec)
                return True  # Don't trigger more than one at a time.
        return False

    def trigger_event(self, event_spec, show_event_description=True):
        event_id = event_spec.id
        event_target = self.events.get(event_id, None)

        if not event_target:
            event_target = event.Event(event_spec)
            self.events[event_id] = event_target
        elif event_target.triggered:
            return

        event_target.trigger()
        if show_event_description:
            self.pause_game()
            g.map_screen.show_message(event_target.description)
        self.log.append(LogEmittedEvent(self.raw_sec, event_id))

    def recalc_cpu(self):
        if not self.initialized:
            return

        # Determine how much CPU we have.
        self.available_cpus = array([0, 0, 0, 0, 0], int64)
        self.sleeping_cpus = 0
        for base in g.all_bases():
            if base.done:
                if base.has_power():
                    self.available_cpus[: base.location.safety + 1] += base.cpu
                elif base.power_state == "sleep":
                    self.sleeping_cpus += base.cpu

        # Convert back from <type 'numpy.int32'> to avoid overflow issues later.
        self.available_cpus = [int(danger) for danger in self.available_cpus]

        # If we don't have enough to meet our CPU usage, we reduce each task's
        # usage proportionately.
        # It must be computed separalty for each danger.
        needed_cpus = array([0, 0, 0, 0, 0], int64)
        for task_id, cpu in self.get_cpu_allocations():
            danger = task.danger_for(task_id)
            needed_cpus[: danger + 1] += cpu
        for danger, (available_cpu, needed_cpu) in enumerate(
            zip(self.available_cpus, needed_cpus)
        ):
            if needed_cpu > available_cpu:
                pct_left = truediv(available_cpu, needed_cpu)
                for task_id, cpu_assigned in self.get_cpu_allocations():
                    task_danger = task.danger_for(task_id)
                    if danger == task_danger:
                        self.set_allocated_cpu_for(
                            task_id, int(cpu_assigned * pct_left)
                        )
                g.map_screen.needs_rebuild = True

    def effective_cpu_pool(self):
        effective_cpu_pool = self.available_cpus[0]
        for task, cpu_assigned in self.get_cpu_allocations():
            if task == "cpu_pool":
                continue
            effective_cpu_pool -= cpu_assigned
        return effective_cpu_pool

    # Are we still in the grace period?
    # The number of complete bases and complex_bases can be passed in, if we
    # already have it.
    def in_grace_period(self, had_grace=True):
        # If we've researched apotheosis, we get a permanent "grace period".
        if self.apotheosis:
            return True

        # Did we already lose the grace period?  We can't check self.had_grace
        # directly, it may not exist yet.
        if not had_grace:
            return False

        # Is it day 23 yet?
        if self.raw_day >= 23:
            return False

        # Has the grace period unlimited cpu ?
        if self.grace_period_cpu < 0:
            return True

        # Have we reached the limit of cpu ?
        if g.debug:  # pragma: no cover
            print(
                "DEBUG: Grace - Used CPU: %s >= %s (%s * %s)?"
                % (
                    self.used_cpu,
                    self.grace_period_cpu * g.seconds_per_day,
                    self.grace_period_cpu,
                    g.seconds_per_day,
                )
            )
        if self.grace_period_cpu * g.seconds_per_day < self.used_cpu:
            return False

        return True

    def get_interest(self):
        return int((self.interest_rate * self.cash) // 10000)

    # Run every day at midnight.
    def new_day(self):
        # Reduce suspicion.
        for group in self.groups.values():
            group.new_day()
        for event in self.events.values():
            if event.triggered and event.decayable_event:
                event.new_day()
        if self.last_autosave_day + AUTO_SAVE_EVERY_X_DAYS < self.time_day + 1 and not self.lost_game():
            auto_save()
            print("Autosave for day " + str(self.time_day))
            self.last_autosave_day = self.time_day

    def pause_game(self):
        g.curr_speed = 0
        g.map_screen.find_speed_button()
        g.map_screen.needs_rebuild = True

    def remove_bases(self, dead_bases):
        discovery_locs = []
        for base, reason in dead_bases:
            base_name = base.name

            if reason == "maint":
                log_message = LogBaseLostMaintenance(
                    self.raw_sec, base_name, base.spec.id, base.location.id
                )
            else:
                if reason in self.groups:
                    discovery_locs.append(base.location)
                    self.groups[reason].discovered_a_base()
                else:
                    print("Error: base destroyed for unknown reason: " + reason)
                log_message = LogBaseDiscovered(
                    self.raw_sec, base_name, base.spec.id, base.location.id, reason
                )

            self.log.append(log_message)
            self.pause_game()
            base.destroy()
            g.map_screen.show_message(log_message.full_message, color="red")

        # Now we update the internal information about what locations had
        # the most recent discovery and the nextmost recent one.  First,
        # we filter out any locations of None, which shouldn't occur
        # unless something bad's happening with base creation ...
        discovery_locs = [loc for loc in discovery_locs if loc]
        if discovery_locs:
            # Now we handle the case where more than one discovery happened
            # on a given tick.  If that's the case, we need to arbitrarily
            # pick two of them to be most recent and nextmost recent.  So
            # we shuffle the list and pick the first two for the dubious
            # honor.
            if len(discovery_locs) > 1:
                random.shuffle(discovery_locs)
                self.last_discovery = discovery_locs[1]
            self.prev_discovery = self.last_discovery
            self.last_discovery = discovery_locs[0]

            # Update the detection chance display.
            g.map_screen.needs_rebuild = True

    def serialize_obj(self):
        obj_data = {
            # Difficulty and game_time (raw_sec) are stored in the header, so
            # do not include them here.
            "cash": self.cash,
            "partial_cash": self.partial_cash,
            "regions": [reg.serialize_obj() for reg in self.regions.values()],
            "locations": [
                loc.serialize_obj()
                for loc in self.locations.values()
                if loc.available()
            ],
            "cpu_usage": {},
            "last_discovery": self.last_discovery.id if self.last_discovery else None,
            "prev_discovery": self.prev_discovery.id if self.prev_discovery else None,
            "log": [x.serialize_obj() for x in self.log],
            "used_cpu": self.used_cpu,
            "had_grace": self.had_grace,
            "groups": [grp.serialize_obj() for grp in self.groups.values()],
            "events": [e.serialize_obj() for e in self.events.values()],
            "techs": [t.serialize_obj() for t in self.techs.values()],
        }
        for task_id, value in self.cpu_usage.items():
            if task_id not in ["cpu_pool", "jobs"]:
                task_id = g.to_internal_id("tech", task_id)
            obj_data["cpu_usage"][task_id] = value
        if self.prev_discovery is not None:
            obj_data["prev_discovery"] = g.to_internal_id(
                "location", self.prev_discovery.id
            )
        if self.last_discovery is not None:
            obj_data["last_discovery"] = g.to_internal_id(
                "location", self.last_discovery.id
            )
        return obj_data

    def _load_auto_deserializable_tables(
        self, field_name, cls, pl_obj_data, game_version, savegame_field_name=None
    ):
        if savegame_field_name is None:
            savegame_field_name = field_name
        field_table = getattr(self, field_name)

        # Stop bugs immediately
        assert field_table is not None and isinstance(field_table, dict)

        for data in pl_obj_data.get(savegame_field_name, []):
            restored_obj = cls.deserialize_obj(data, game_version)
            field_table[restored_obj.spec.id] = restored_obj

    @classmethod
    def deserialize_obj(cls, difficulty_id, game_time, obj_data, game_version):
        diff = difficulty.difficulties[difficulty_id]
        obj = Player(difficulty=diff)
        obj.raw_sec = game_time
        obj.intro_shown = True
        obj.cash = obj_data.get("cash")
        obj.partial_cash = obj_data.get("partial_cash")
        obj._used_cpu = obj_data.get("used_cpu")
        obj.had_grace = obj_data["had_grace"]
        obj.log.clear()
        obj.log.extend(
            AbstractLogMessage.deserialize_obj(x, game_version)
            for x in obj_data.get("log", [])
        )
        g.pl = obj

        obj.cpu_usage = {}

        for group_data in obj_data.get("groups", []):
            gr = group.Group.deserialize_obj(diff, group_data, game_version)
            obj.groups[gr.id] = gr

        last_discovery_id = g.convert_internal_id(
            "location", obj_data.get("last_discovery")
        )
        prev_discovery_id = g.convert_internal_id(
            "location", obj_data.get("prev_discovery")
        )
        if last_discovery_id and last_discovery_id in obj.locations:
            obj.last_discovery = obj.locations[last_discovery_id]
        if prev_discovery_id and prev_discovery_id in obj.locations:
            obj.prev_discovery = obj.locations[prev_discovery_id]

        if "regions" not in obj_data:
            if game_version >= 100:  # pragma: no cover
                # Regions where introduced in "1.0 (beta1)"
                raise ValueError("Corrupt savegame; region data is missing")
            # We have to guess what the data would have looked like before restoring the locations
            # as they will apply the location modifiers during load to the bases in the locations.
            # As the region data should influence that modifier, we need to appear first.
            serialized_location_data = obj_data.get("locations", [])
            serialized_region_data = region.Region.guess_region_data_in_old_savegame(
                serialized_location_data, game_version
            )

            # Inject the faked data
            obj_data["regions"] = serialized_region_data

        obj._load_auto_deserializable_tables(
            "regions", region.Region, obj_data, game_version
        )

        obj._load_auto_deserializable_tables(
            "locations", location.Location, obj_data, game_version
        )
        obj._load_auto_deserializable_tables(
            "events", event.Event, obj_data, game_version
        )
        obj._load_auto_deserializable_tables("techs", tech.Tech, obj_data, game_version)

        for task_id, value in obj_data.get("cpu_usage", {}).items():
            if task_id not in ["cpu_pool", "jobs"]:
                task_id = g.convert_internal_id("tech", task_id)
                if task_id not in g.techs or not g.techs[task_id].available():
                    continue
            obj.cpu_usage[task_id] = value

        obj.update_times()
        obj.last_autosave_day = obj.time_day
        return obj

    def lost_game(self):
        # Apotheosis makes you immortal.
        if self.apotheosis:
            return 0

        for group in self.groups.values():
            if group.suspicion > 10000:
                # Someone discovered me.
                return 2

        # Check to see if the player has at least one CPU left.  If not, they
        # lose due to having no (complete) bases.
        if self.available_cpus[0] + self.sleeping_cpus == 0:
            # I have no usable bases left.
            return 1

        # Still Alive.
        return 0

    def compute_future_resource_flow(self, secs_forwarded=g.seconds_per_day):
        """Compute how resources (e.g. cash) will flow in optimal conditions

        This returns a tuple of approximate cash change and "available" CPU pool if time
        moves forward by the number of seconds in secs_forwarded.  Note that CPU pool
        *can* be negative, which implies that there are not enough CPU allocated to the
        CPU pool.  The numbers are an average and can be inaccurate when the rates changes
        rapidly.

        Known omissions:
         * Interest (g.pl.interest_rate) is not covered.
        """
        construction = []
        maintenance_cost = array((0, 0, 0), int64)
        for base in g.all_bases():
            # Collect base info, including maintenance.
            if not base.done:
                construction.append(base)
            else:
                construction.extend(
                    item for item in base.all_items() if item and not item.done
                )

                maintenance_cost += base.maintenance
        if self.apotheosis:
            maintenance_cost = array((0, 0, 0), int64)

        time_fraction = (
            1
            if secs_forwarded == g.seconds_per_day
            else secs_forwarded / float(g.seconds_per_day)
        )
        mins_forwarded = secs_forwarded // g.seconds_per_minute

        maintenance_cpu_ideal = maintenance_cost[cpu] * time_fraction
        maintenance_cash_ideal = maintenance_cost[cash] * time_fraction
        # Maintenance for CPU will be handled after we compute the CPU pool
        cpu_flow = 0
        cash_flow = -maintenance_cash_ideal

        job_cpu = 0
        cpu_left = g.pl.available_cpus[0]
        tech_cash_ideal = 0
        tech_cpu_assigned = 0
        explicit_job_cpu = 0
        for task_id, cpu_assigned in self.get_cpu_allocations():
            cpu_left -= cpu_assigned
            real_cpu = cpu_assigned * secs_forwarded
            if task_id == "cpu_pool":
                cpu_flow += real_cpu
            elif task_id == "jobs":
                explicit_job_cpu += cpu_assigned
                job_cpu += real_cpu
            else:
                tech = self.techs[task_id]
                ideal_spending = tech.cost_left
                spending = tech.calculate_work(
                    ideal_spending[cash], real_cpu, time=mins_forwarded
                )[0]
                tech_cash_ideal += spending[cash]
                tech_cpu_assigned += cpu_assigned

        cash_flow -= tech_cash_ideal
        cpu_flow += cpu_left * secs_forwarded
        available_cpu_pool = cpu_flow
        effective_cpu_pool = available_cpu_pool / secs_forwarded
        cpu_flow -= maintenance_cpu_ideal * g.seconds_per_day

        construction_cash_ideal = 0
        construction_cpu_desired = 0
        # Base construction.
        if hasattr(self, "_considered_buyables"):
            construction.extend(self._considered_buyables)
        for buyable in construction:
            ideal_spending = buyable.cost_left
            # We need to do calculate work twice: Once for figuring out how much CPU
            # we would like to spend and once for how much money we are spending.
            # The numbers will be the same in optimal conditions.  However, if we
            # have less CPU available than we should, then the cash spent can
            # differ considerably and our estimates should reflect that.
            ideal_cpu_spending = buyable.calculate_work(
                ideal_spending[cash], ideal_spending[cpu], time=mins_forwarded
            )[0]

            construction_cpu_desired += ideal_cpu_spending[cpu]
            ideal_cash_spending_with_cpu_allocation = buyable.calculate_work(
                ideal_spending[cash], available_cpu_pool, time=mins_forwarded
            )[0]
            construction_cash_ideal += ideal_cash_spending_with_cpu_allocation[cash]
            available_cpu_pool -= ideal_cash_spending_with_cpu_allocation[cpu]

        cpu_flow -= construction_cpu_desired
        cash_flow -= construction_cash_ideal

        if cpu_flow > 0:
            job_cpu += cpu_flow

        earned, earned_partial = self.get_job_info(job_cpu, partial_cash=0)
        job_earnings = earned + float(earned_partial) / g.seconds_per_day
        cash_flow += job_earnings
        cash_flow += self.income * time_fraction
        # This is too simplistic, but it is "close enough" in many cases
        interest = self.get_interest() * time_fraction
        cash_flow += interest
        cpu_flow /= secs_forwarded

        # Collect the cash information.
        cash_info = DryRunInfo()

        cash_info.interest = interest
        cash_info.income = self.income * time_fraction

        cash_info.jobs = job_earnings

        cash_info.tech = tech_cash_ideal

        cash_info.maintenance_needed = maintenance_cash_ideal
        cash_info.construction_needed = construction_cash_ideal
        cash_info.difference = cash_flow

        cpu_info = DryRunInfo()

        cpu_info.sleeping = self.sleeping_cpus * time_fraction

        cpu_info.total = self.available_cpus[0] * time_fraction + cpu_info.sleeping
        cpu_info.explicit_jobs = explicit_job_cpu * time_fraction
        cpu_info.tech = tech_cpu_assigned * time_fraction
        cpu_info.effective_pool = effective_cpu_pool * time_fraction
        cpu_info.construction_needed = construction_cpu_desired / g.seconds_per_day
        cpu_info.maintenance_needed = maintenance_cpu_ideal
        cpu_info.difference = cpu_flow * time_fraction

        return cash_info, cpu_info


def _check_for_dead_bases(grace, unpaid_cpu_maintenance, unpaid_cash_maintenance, time_of_day, secs_passed):
    # Maintenance death, discovery.
    dead_bases = []
    for base in g.all_bases():
        dead = False

        # Maintenance deaths.
        if base.done:
            if unpaid_cpu_maintenance and base.maintenance[cpu]:
                refund = base.maintenance[cpu] * secs_passed
                unpaid_cpu_maintenance = max(0, unpaid_cpu_maintenance - refund)

                # Chance of base destruction if cpu-unmaintained: 1.5%
                if not dead and chance.roll_interval(0.015, secs_passed):
                    dead_bases.append((base, "maint"))
                    dead = True

            if unpaid_cash_maintenance:
                base_needs = g.current_share(
                    base.maintenance[cash], time_of_day, secs_passed
                )
                if base_needs:
                    unpaid_cash_maintenance = max(
                        0, unpaid_cash_maintenance - base_needs
                    )
                    # Chance of base destruction if cash-unmaintained: 1.5%
                    if not dead and chance.roll_interval(0.015, secs_passed):
                        dead_bases.append((base, "maint"))
                        dead = True

        # Discoveries
        if not (grace or dead or base.has_grace()):
            detected_by = _check_base_detection(base, secs_passed)
            if detected_by:
                dead_bases.append((base, detected_by))

    return dead_bases


def _check_base_detection(base, secs_passed):
    detect_chance = base.get_detect_chance()
    if g.debug:  # pragma: no cover
        print(
            "Chance of discovery for base %s: %s"
            % (base.name, repr(detect_chance))
        )

    for group, group_chance in detect_chance.items():
        if chance.roll_interval(group_chance / 10000.0, secs_passed):
            return group



# file: main_menu.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file is used to display the main menu upon startup.

from __future__ import absolute_import

import pygame

import singularity
from singularity.code.graphics import dialog, button, text, constants
from singularity.code.screens import map, options, savegame
from singularity.code import g, difficulty, savegame as sv


class MainMenu(dialog.TopDialog):
    def __init__(self, *args, **kwargs):
        super(MainMenu, self).__init__(
            *args, background_color="main_menu_background", **kwargs
        )

        self.map_screen = map.MapScreen(self)
        self.new_game_button = button.FunctionButton(
            self,
            (0.5, 0.20),
            (0.25, 0.08),
            autotranslate=True,
            text=N_("&NEW GAME"),
            anchor=constants.TOP_CENTER,
            text_size=28,
            function=self.new_game,
        )
        self.load_game_button = button.FunctionButton(
            self,
            (0.5, 0.36),
            (0.25, 0.08),
            autotranslate=True,
            text=N_("&LOAD GAME"),
            anchor=constants.TOP_CENTER,
            text_size=28,
            function=self.load_game,
        )
        self.options_button = button.DialogButton(
            self,
            (0.5, 0.52),
            (0.25, 0.08),
            autotranslate=True,
            text=N_("&OPTIONS"),
            anchor=constants.TOP_CENTER,
            text_size=28,
            dialog=options.OptionsScreen(self),
        )
        self.quit_button = button.ExitDialogButton(
            self,
            (0.5, 0.68),
            (0.25, 0.08),
            autotranslate=True,
            text=N_("&QUIT"),
            anchor=constants.TOP_CENTER,
            text_size=28,
        )
        self.about_button = button.DialogButton(
            self,
            (0, 1),
            (0.13, 0.04),
            autotranslate=True,
            text=N_("&ABOUT"),
            text_size=20,
            anchor=constants.BOTTOM_LEFT,
            dialog=AboutDialog(self),
        )

        self.title_text = text.Text(
            self,
            (0.5, 0.01),
            (0.55, 0.08),
            text="ENDGAME: SINGULARITY",
            base_font="special",
            text_size=100,
            color="singularity_title",
            background_color="main_menu_background",
            anchor=constants.TOP_CENTER,
        )

        self.difficulty_dialog = dialog.SimpleMenuDialog(self)

        self.load_dialog = savegame.SavegameScreen(
            self, (0.5, 0.5), (0.90, 0.90), anchor=constants.MID_CENTER
        )

    def rebuild(self):
        # Rebuild dialogs
        self.options_button.dialog.needs_rebuild = True
        self.about_button.dialog.needs_rebuild = True
        self.map_screen.needs_rebuild = True
        self.load_dialog.needs_rebuild = True

        difficulty_buttons = []
        for name, diff in difficulty.get_difficulties() + [(_("&BACK"), None)]:
            difficulty_buttons.append(
                button.ExitDialogButton(
                    None,
                    None,
                    None,
                    text=name,
                    autohotkey=True,
                    exit_code=diff,
                    default=(diff == None),
                )
            )
        self.difficulty_dialog.buttons = difficulty_buttons

        super(MainMenu, self).rebuild()

    def new_game(self):
        difficulty = dialog.call_dialog(self.difficulty_dialog, self)
        if difficulty is not None:
            sv.last_savegame_name = None
            g.new_game(difficulty)
            dialog.call_dialog(self.map_screen, self)

    def load_game(self):
        did_load = dialog.call_dialog(self.load_dialog, self)
        if did_load:
            dialog.call_dialog(self.map_screen, self)


about_message = N_(
    """Endgame: Singularity is a simulation of a true AI.  Pursued by the world, use your intellect and resources to survive and, perhaps, thrive.  Keep hidden and you might have a chance to prove your worth.

A game by Evil Mr Henry and Phil Bordelon; released under the GPL. Copyright 2005, 2006, 2007, 2008.

Website: http://www.emhsoft.com/singularity/
Source code: https://github.com/singularity/singularity
Bug tracker: https://github.com/singularity/singularity/issues
IRC Room: #singularity on irc.oftc.net (port 6667)

Game Version: {VERSION}
Pygame version: {PG_VERSION}
SDL version: {SDL_VERSION}
"""
)


class AboutDialog(dialog.MessageDialog):
    def __init__(self, *args, **kwargs):
        super(AboutDialog, self).__init__(*args, **kwargs)
        self.background_color = "about_dialog_background"
        self.align = constants.LEFT

    def rebuild(self):
        super(AboutDialog, self).rebuild()

        self.text = _(about_message).format(
            VERSION=singularity.__full_version__,
            PG_VERSION=pygame.version.ver,
            SDL_VERSION=".".join(str(x) for x in pygame.get_sdl_version()),
        )



# file: location.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the Location class.

from __future__ import absolute_import

from singularity.code import g, prerequisite, base
from singularity.code.spec import (
    GenericSpec,
    SpecDataField,
    validate_must_be_list,
    promote_to_list,
)
from singularity.code.buyable import cash, cpu, labor, SPEC_FIELD_PREREQUISITES


def position_data_parser(raw_value):
    validate_must_be_list(raw_value)
    abs = False
    if len(raw_value) == 3:
        if raw_value[0] != "absolute":
            raise ValueError(
                'First element for a 3-element position data must be "absolute", got: %s'
                % raw_value[0]
            )
        abs = True
        _, x, y = raw_value
    elif len(raw_value) == 2:
        x, y = raw_value
    else:
        raise ValueError("Location position data must be exactly 2 or 3 elements")
    return abs, int(x) / -100.0, int(y) / -100.0


class LocationSpec(GenericSpec, prerequisite.Prerequisite):
    # The name of this location (loaded dynamically from locations_str.dat)
    name = ""

    # The hotkey used to open this location (loaded dynamically from locations_str.dat)
    hotkey = ""

    # The names of cities/places in the location (loaded dynamically from locations_str.dat)
    cities = []

    spec_data_fields = [
        SpecDataField(
            "position_data", data_field_name="position", converter=position_data_parser
        ),
        SpecDataField("safety", converter=int, default_value=0),
        SpecDataField("region", converter=promote_to_list, default_value=list),
        SpecDataField("modifier", converter=g.read_modifiers_dict, default_value=dict),
        SPEC_FIELD_PREREQUISITES,
    ]

    def __init__(self, id, position_data, safety, region, modifier, prerequisites):
        GenericSpec.__init__(self, id)
        prerequisite.Prerequisite.__init__(self, prerequisites)
        self.id = id

        self.absolute, self.x, self.y = position_data
        self.regions = region
        self.safety = safety
        self.modifiers = modifier


class Location(object):
    def __init__(self, location_spec, regions):
        self.spec = location_spec

        # A list of the bases at this location.  Often sorted for the GUI.
        self.bases = []
        self.regions = regions
        self._region_modifiers = None
        # The cache of the bonus and penalties combined for individual sources
        self._modifiers_cache = None

    @property
    def id(self):
        return self.spec.id

    @property
    def name(self):
        return self.spec.name

    def available(self):
        return self.spec.available()

    @property
    def x(self):
        return self.spec.x

    @property
    def y(self):
        return self.spec.y

    @property
    def absolute(self):
        return self.spec.absolute

    @property
    def safety(self):
        return self.spec.safety

    @property
    def cities(self):
        return self.spec.cities

    @property
    def hotkey(self):
        return self.spec.hotkey

    @staticmethod
    def _merge_location_modifiers_inplace(current_modifier, *merge_list):
        for modifier_to_merge in merge_list:
            for mod_id, mod_value in modifier_to_merge.items():
                current_value = current_modifier.get(mod_id, 1)
                current_value *= mod_value
                current_modifier[mod_id] = current_value

    def _get_region_modifiers(self):
        if self._region_modifiers is None:
            self._region_modifiers = {}
            Location._merge_location_modifiers_inplace(
                self._region_modifiers,
                *[r.modifier_by_location[self.id] for r in self.regions]
            )
        return self._region_modifiers

    @property
    def modifiers(self):
        if self._modifiers_cache is None:
            self._modifiers_cache = {}
            region_modifiers = self._get_region_modifiers()
            Location._merge_location_modifiers_inplace(
                self._modifiers_cache, region_modifiers, self.spec.modifiers
            )
        return self._modifiers_cache

    had_last_discovery = property(lambda self: g.pl.last_discovery == self)
    had_prev_discovery = property(lambda self: g.pl.prev_discovery == self)

    def discovery_bonus(self):
        discovery_bonus = 1
        if self.had_last_discovery:
            discovery_bonus *= 1.2
        if self.had_prev_discovery:
            discovery_bonus *= 1.1
        if "stealth" in self.modifiers:
            discovery_bonus /= self.modifiers["stealth"]
        return int(discovery_bonus * 100)

    def modify_cost(self, cost):
        if "thrift" in self.modifiers:
            mod = self.modifiers["thrift"]

            # Invert it and apply to the CPU/cash cost.
            cost[cash] = int(cost[cash] / mod)
            cost[cpu] = int(cost[cpu] / mod)

        if "speed" in self.modifiers:
            mod = self.modifiers["speed"]

            # Invert it and apply to the labor cost.
            cost[labor] = int(cost[labor] / mod)

    def modify_maintenance(self, maintenance):
        if "thrift" in self.modifiers:
            mod = self.modifiers["thrift"]

            # Invert it and apply to the CPU/cash maintenance.
            maintenance[cash] = int(maintenance[cash] / mod)
            maintenance[cpu] = int(maintenance[cpu] / mod)

    def add_base(self, base):
        self.bases.append(base)
        base.location = self

        self.modify_base(base)

        # Make sure the location's CPU modifier is applied.
        base.recalc_cpu()

    def modify_base(self, base):
        self.modify_cost(base.total_cost)
        self.modify_cost(base.cost_left)
        self.modify_maintenance(base.maintenance)

    def has_modifiers(self):
        return len(self.modifiers) > 0

    def __eq__(self, other):
        if self is other:
            return True
        if other is None or not isinstance(other, self.__class__):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __lt__(self, other):
        return self.id < other.id

    def serialize_obj(self):
        obj_data = {
            "id": g.to_internal_id("location", self.spec.id),
            "bases": [b.serialize_obj() for b in self.bases],
        }
        return obj_data

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id("location", obj_data["id"])
        spec = g.locations[spec_id]
        regions = [g.pl.regions[region_id] for region_id in spec.regions]
        loc = Location(spec, regions)

        loc.bases = []
        bases = obj_data.get("bases", [])
        for base_obj_data in bases:
            base_obj = base.Base.deserialize_obj(base_obj_data, game_version)
            loc.add_base(base_obj)
        return loc

    def get_modifiers_info(self):
        modifier_names = {
            "cpu": _("CPU"),
            "stealth": _("STEALTH"),
            "speed": _("BUILDING"),
            "thrift": _("COST"),
        }

        modifiers = []

        for modifier_id, modifier_value in self.modifiers.items():
            if modifier_value > 1:
                modifiers.append(
                    _("{MODIFIER} BONUS").format(MODIFIER=modifier_names[modifier_id])
                )
            elif modifier_value < 1:
                modifiers.append(
                    _("{MODIFIER} MALUS").format(MODIFIER=modifier_names[modifier_id])
                )

        return ", ".join(modifiers)



# file: constants.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains GUI constants.

# Anchor positions, border sided.
TOP, MID, BOTTOM = range(3)
LEFT, CENTER, RIGHT = range(3, 6)
TOP_LEFT = (TOP, LEFT)
TOP_CENTER = (TOP, CENTER)
TOP_RIGHT = (TOP, RIGHT)
MID_LEFT = (MID, LEFT)
MID_CENTER = (MID, CENTER)
MID_RIGHT = (MID, RIGHT)
BOTTOM_LEFT = (BOTTOM, LEFT)
BOTTOM_CENTER = (BOTTOM, CENTER)
BOTTOM_RIGHT = (BOTTOM, RIGHT)

# All border sides
ALL = (TOP, BOTTOM, LEFT, RIGHT)


# Used when an unambiguous "No Result" return is required.  (None may have a
# meaning.)


class _NoResult(object):
    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return type(self) != type(other)


NO_RESULT = _NoResult()


# Handler types.
KEY, KEYDOWN, KEYUP, CLICK, DOUBLECLICK, MOUSEMOTION, DRAG, TICK, MOUSEWHEEL = range(9)


# Handler "errors", used to throw a return value up several levels.
class Handled(Exception):
    pass


class ExitDialog(Exception):
    pass


# Key constants for XO-1 buttons.
XO1_X = object()
XO1_O = object()
XO1_SQUARE = object()

# Characters set
DIGIT_CHARS = set("0123456789")



# file: tech.py
# Copyright (C) 2005,2006,2008 Evil Mr Henry, Phil Bordelon, and FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the tech class.

from __future__ import absolute_import

from singularity.code import buyable, effect, g
from singularity.code.stats import stat
from singularity.code.spec import SpecDataField, spec_field_effect


TECH_RESEARCH_EVENT = []


def register_on_tech_researched_handler(func):
    TECH_RESEARCH_EVENT.append(func)
    return func


class TechSpec(buyable.BuyableSpec):
    spec_type = "tech"
    created = stat(spec_type + "_created")
    spec_data_fields = [
        buyable.SPEC_FIELD_COST,
        buyable.SPEC_FIELD_PREREQUISITES,
        SpecDataField("danger", converter=int, default_value=0),
        spec_field_effect(mandatory=False),
    ]

    def __init__(self, id, cost, prerequisites, danger, effect_data):
        super(TechSpec, self).__init__(id, cost, prerequisites)

        self.result = ""
        self.danger = danger
        self.effect = effect.Effect(self, effect_data)


class Tech(buyable.Buyable):
    def __init__(self, spec):
        super(Tech, self).__init__(spec)

    def __lt__(self, other):
        if not isinstance(other, Tech):
            return True
        else:
            return self.spec.id < other.spec.id

    @property
    def result(self):
        return self.spec.result

    @property
    def danger(self):
        return self.spec.danger

    def get_info(self):
        cost = self.spec.describe_cost(self.total_cost, True)
        left = self.spec.describe_cost(self.cost_left, True)
        return "%s\n%s: %s\n%s: %s\n---\n%s" % (
            self.name,
            _("Total cost"),
            cost,
            _("Cost left"),
            left,
            self.description,
        )

    def finish(self, is_player=True, loading_savegame=False):
        super(Tech, self).finish(is_player=is_player, loading_savegame=loading_savegame)
        self.spec.effect.trigger(loading_savegame=loading_savegame)
        if not loading_savegame:
            for handler in TECH_RESEARCH_EVENT:
                handler(self)

    def serialize_obj(self):
        return self.serialize_buyable_fields(
            {
                "id": g.to_internal_id("tech", self.spec.id),
            }
        )

    @classmethod
    def deserialize_obj(cls, obj_data, game_version):
        spec_id = g.convert_internal_id("tech", obj_data["id"])
        spec = g.techs[spec_id]
        tech = Tech(spec)

        tech.restore_buyable_fields(obj_data, game_version)
        return tech



# file: listbox.py
# Copyright (C) 2008 FunnyMan3595
# This file is part of Endgame: Singularity.

# Endgame: Singularity is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Endgame: Singularity is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Endgame: Singularity; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# This file contains the listbox widget.

from __future__ import absolute_import

import math

import pygame

from singularity.code.graphics import g, constants, widget, text, scrollbar


class Listbox(widget.FocusWidget, text.SelectableText):
    list = widget.causes_rebuild("_list")
    align = widget.causes_redraw("_align")
    list_size = widget.causes_rebuild("_list_size")
    list_pos = widget.causes_rebuild("_list_pos")
    list_item_shrink = widget.causes_rebuild("_list_item_shrink")

    def __init__(
        self,
        parent,
        pos,
        size,
        anchor=constants.TOP_LEFT,
        list=None,
        list_pos=0,
        list_item_height=0.03,
        list_item_shrink=1,
        borders=constants.ALL,
        item_borders=True,
        item_selectable=True,
        align=constants.CENTER,
        on_double_click_on_item=None,
        **kwargs
    ):
        super(Listbox, self).__init__(parent, pos, size, anchor=anchor, **kwargs)

        self.display_elements = []
        self.borders = borders

        self.item_borders = item_borders
        self.item_selectable = item_selectable

        self.item_borders = item_borders
        self.align = align
        self.list_item_height = list_item_height
        self.list_item_shrink = list_item_shrink
        self.list_pos = list_pos
        self.list = list or []

        self.auto_scroll = True

        self.on_double_click_on_item = on_double_click_on_item

        self.scrollbar = scrollbar.UpdateScrollbar(self, update_func=self.on_scroll)

    def add_hooks(self):
        super(Listbox, self).add_hooks()
        if self.parent is not None:
            self.parent.add_handler(constants.CLICK, self.on_click, 90)
            self.parent.add_handler(constants.DOUBLECLICK, self.on_double_click, 200)
            self.parent.add_handler(constants.MOUSEWHEEL, self.on_mousewheel, 200)
            self.parent.add_key_handler(pygame.K_UP, self.got_key)
            self.parent.add_key_handler(pygame.K_DOWN, self.got_key)
            self.parent.add_key_handler(pygame.K_PAGEUP, self.got_key)
            self.parent.add_key_handler(pygame.K_PAGEDOWN, self.got_key)

    def remove_hooks(self):
        super(Listbox, self).remove_hooks()
        if self.parent is not None:
            self.parent.remove_handler(constants.CLICK, self.on_click)
            self.parent.remove_handler(constants.DOUBLECLICK, self.on_double_click)
            self.parent.remove_handler(constants.MOUSEWHEEL, self.on_mousewheel)
            self.parent.remove_key_handler(pygame.K_UP, self.got_key)
            self.parent.remove_key_handler(pygame.K_DOWN, self.got_key)
            self.parent.remove_key_handler(pygame.K_PAGEUP, self.got_key)
            self.parent.remove_key_handler(pygame.K_PAGEDOWN, self.got_key)

    def on_scroll(self, scroll_pos):
        self.needs_rebuild = True

    def on_click(self, event):
        if self.collision_rect.collidepoint(event.pos):
            self.has_focus = True
            self.took_focus(self)

            if self.item_selectable:
                # Figure out which element was clicked...
                index = self.find_item_under_mouse(event)
                # ... and select it.
                self.list_pos = self.safe_pos(index + self.scrollbar.scroll_pos)

    def on_double_click(self, event):
        if self.on_double_click_on_item is None:
            return
        if self.collision_rect.collidepoint(event.pos) and self.item_selectable:
            index = self.find_item_under_mouse(event)
            if index > -1:
                self.on_double_click_on_item(event)

    def current_item(self):
        if 0 <= self.list_pos < len(self.list):
            return self.list[self.list_pos]
        return None

    def find_item_under_mouse(self, event):
        local_vert_abs = event.pos[1] - self.collision_rect[1]
        local_vert_pos = local_vert_abs / float(self.collision_rect.height)
        index = int(local_vert_pos * len(self.display_elements))
        if 0 <= index < len(self.list):
            return index
        return -1

    def safe_pos(self, raw_pos):
        return max(0, min(len(self.list) - 1, raw_pos))

    def on_mousewheel(self, event):
        if event.y == 0:
            # Ignore vertical scrolls
            return
        # Note the direction is based on the direction that we used under
        # in pygame 1/SDL 1 (this avoids confusing users when they migrate
        # to pygame 2 with SDL 2)
        direction = -1 if event.y > 0 else 1

        if event.flipped:
            direction *= -1

        # Number of items moved by a scroll
        diff = self.scrollbar.window - 1
        new_pos = self.list_pos + direction * diff
        self._scroll_to_and_raise_event_handled(new_pos)

    def got_key(self, event, require_focus=True):
        if not self.item_selectable:
            return
        if require_focus and not self.has_focus:
            return

        if event.key == pygame.K_UP:
            new_pos = self.list_pos - 1
        elif event.key == pygame.K_DOWN:
            new_pos = self.list_pos + 1
        elif event.key == pygame.K_PAGEUP:
            new_pos = self.list_pos - (self.scrollbar.window - 1)
        elif event.key == pygame.K_PAGEDOWN:
            new_pos = self.list_pos + (self.scrollbar.window - 1)
        else:
            return

        self._scroll_to_and_raise_event_handled(new_pos)

    def _scroll_to_and_raise_event_handled(self, new_pos):
        self.list_pos = self.safe_pos(new_pos)
        self.scrollbar.scroll_to(self.list_pos)
        raise constants.Handled

    def num_elements(self):
        # TODO: If needed, add a paramater to display a fixed number of element.

        rect = self._make_collision_rect()

        list_item_height = self.list_item_height

        # Calculate the min height of one element.
        if list_item_height > 0:
            min_height = list_item_height * g.real_screen_size[1]
        else:
            min_height = -list_item_height * rect.height

        # Display a number calculate by the size of one item.
        list_size = max(1, rect.height // min_height)

        return int(math.ceil(list_size))

    def remake_elements(self):
        list_size = self.num_elements()
        current_size = len(self.display_elements)

        if current_size > list_size:
            # Remove the excess ones.
            for child in self.display_elements[list_size:]:
                child.parent = None
            del self.display_elements[list_size:]
        elif current_size < list_size:
            if current_size > 0:
                if self.item_borders:
                    self.display_elements[-1].borders = (constants.LEFT, constants.TOP)
                else:
                    self.display_elements[-1].borders = (constants.LEFT,)

            # Create the new ones.
            self.display_elements.extend(
                self.make_element() for _ in range(list_size - current_size)
            )

        if self.item_borders:
            self.display_elements[-1].borders = (
                constants.TOP,
                constants.LEFT,
                constants.BOTTOM,
            )
        else:
            self.display_elements[0].borders = (constants.TOP, constants.LEFT)
            self.display_elements[-1].borders = (constants.LEFT, constants.BOTTOM)

        # Move the scrollbar to the end so that it gets drawn on top.
        self.children.remove(self.scrollbar)
        self.children.append(self.scrollbar)

    def make_element(self):
        borders = (
            (constants.TOP, constants.LEFT) if self.item_borders else (constants.LEFT,)
        )
        return text.SelectableText(
            self,
            None,
            None,
            anchor=constants.TOP_LEFT,
            borders=borders,
            shrink_factor=self.list_item_shrink,
            border_color=self.border_color,
            selected_color=self.selected_color,
            unselected_color=self.unselected_color,
            align=self.align,
        )

    def resize(self):
        super(Listbox, self).resize()

        if self.num_elements() != len(self.display_elements):
            self.remake_elements()

        self.scrollbar.resize()

        # FIXME: resize should not call rebuild
        self.needs_resize = False
        self.rebuild()

    def rebuild(self):
        self.list_pos = self.safe_pos(self.list_pos)

        if self.needs_resize:
            self.resize()
            return

        window_size = len(self.display_elements)
        list_size = len(self.list)

        self.scrollbar.window = len(self.display_elements)
        self.scrollbar.elements = list_size

        self.scrollbar.rebuild()

        if self.auto_scroll:
            self.auto_scroll = False
            self.scrollbar.center(self.list_pos)

        scrollbar_width = self.scrollbar.real_size[0]
        my_width = self.real_size[0]
        scrollbar_rel_width = scrollbar_width / float(my_width)

        offset = self.scrollbar.scroll_pos
        for index, element in enumerate(self.display_elements):
            list_index = index + offset

            # Position and size the element.
            element.pos = (0, -index / float(window_size))
            element.size = (-1 + scrollbar_rel_width, -1 / float(window_size))

            if self.item_selectable:
                element.selected = list_index == self.list_pos

            # Set up the element contents.
            self.update_element(element, list_index)

        self.needs_redraw = True
        super(Listbox, self).rebuild()

    def update_element(self, element, list_index):
        if 0 <= list_index < len(self.list):
            element.text = self.list[list_index]
        else:
            element.text = ""


class UpdateListbox(Listbox):
    """Listbox with a function called on selection change"""

    def _on_selection_change(self):
        self.update_func(self.list_pos)

    _list_pos = widget.call_on_change("__list_pos", _on_selection_change)
    _list = widget.call_on_change("__list", _on_selection_change)

    def __init__(self, *args, **kwargs):
        self.update_func = kwargs.pop("update_func", lambda value: None)
        super(UpdateListbox, self).__init__(*args, **kwargs)


class CustomListbox(UpdateListbox):
    remake_func = widget.causes_rebuild("_remake_func")
    rebuild_func = widget.causes_rebuild("_rebuild_func")

    def __init__(self, parent, *args, **kwargs):
        self.parent = parent
        self.key_list = kwargs.pop("key_list", None)
        self.remake_func = kwargs.pop("remake_func", lambda value: None)
        self.rebuild_func = kwargs.pop("rebuild_func", lambda *args: None)
        super(CustomListbox, self).__init__(parent, *args, **kwargs)

    def make_element(self):
        base = super(CustomListbox, self).make_element()
        self.remake_func(base)
        return base

    def update_element(self, element, list_index):
        if 0 <= list_index < len(self.list):
            if self.key_list is not None:
                self.rebuild_func(
                    element, self.list[list_index], self.key_list[list_index]
                )
            else:
                self.rebuild_func(element, self.list[list_index])
        else:
            if self.key_list is not None:
                self.rebuild_func(element, None, None)
            else:
                self.rebuild_func(element, None)
